from collections import defaultdict
from os import path
from unittest import TestCase
from unittest.mock import mock_open, patch

import pyinfra
from pyinfra.api import (
    BaseStateCallback,
    Config,
    FileDownloadCommand,
    FileUploadCommand,
    OperationError,
    OperationValueError,
    State,
    StringCommand,
)
from pyinfra.api.connect import connect_all, disconnect_all
from pyinfra.api.exceptions import PyinfraError
from pyinfra.api.operation import OperationMeta, add_op
from pyinfra.api.operations import run_ops
from pyinfra.context import ctx_host, ctx_state
from pyinfra.operations import files, python, server

from ..paramiko_util import FakeBuffer, FakeChannel, PatchSSHTestCase
from ..util import make_inventory


class TestOperationMeta(TestCase):
    def test_operation_meta_repr_no_change(self):
        op_meta = OperationMeta("hash", [])
        assert repr(op_meta) == "OperationMeta(commands=0, changed=False, hash=hash)"

    def test_operation_meta_repr_changes(self):
        op_meta = OperationMeta("hash", ["a-command"])
        assert repr(op_meta) == "OperationMeta(commands=1, changed=True, hash=hash)"


class TestOperationsApi(PatchSSHTestCase):
    def test_op(self):
        inventory = make_inventory()
        somehost = inventory.get_host("somehost")
        anotherhost = inventory.get_host("anotherhost")

        state = State(inventory, Config())
        state.add_callback_handler(BaseStateCallback())

        # Enable printing on this test to catch any exceptions in the formatting
        state.print_output = True
        state.print_input = True
        state.print_fact_info = True
        state.print_noop_info = True

        connect_all(state)

        add_op(
            state,
            files.file,
            "/var/log/pyinfra.log",
            user="pyinfra",
            group="pyinfra",
            mode="644",
            create_remote_dir=False,
            sudo=True,
            sudo_user="test_sudo",
            su_user="test_su",
            ignore_errors=True,
            env={
                "TEST": "what",
            },
        )

        op_order = state.get_op_order()

        # Ensure we have an op
        assert len(op_order) == 1

        first_op_hash = op_order[0]

        # Ensure the op name
        assert state.op_meta[first_op_hash]["names"] == {"Files/File"}

        # Ensure the commands
        assert state.ops[somehost][first_op_hash]["commands"] == [
            StringCommand("touch /var/log/pyinfra.log"),
            StringCommand("chmod 644 /var/log/pyinfra.log"),
            StringCommand("chown pyinfra:pyinfra /var/log/pyinfra.log"),
        ]

        # Ensure the global kwargs (same for both hosts)
        somehost_global_kwargs = state.ops[somehost][first_op_hash]["global_kwargs"]
        assert somehost_global_kwargs["sudo"] is True
        assert somehost_global_kwargs["sudo_user"] == "test_sudo"
        assert somehost_global_kwargs["su_user"] == "test_su"
        assert somehost_global_kwargs["ignore_errors"] is True

        anotherhost_global_kwargs = state.ops[anotherhost][first_op_hash]["global_kwargs"]
        assert anotherhost_global_kwargs["sudo"] is True
        assert anotherhost_global_kwargs["sudo_user"] == "test_sudo"
        assert anotherhost_global_kwargs["su_user"] == "test_su"
        assert anotherhost_global_kwargs["ignore_errors"] is True

        # Ensure run ops works
        run_ops(state)

        # Ensure ops completed OK
        assert state.results[somehost]["success_ops"] == 1
        assert state.results[somehost]["ops"] == 1
        assert state.results[anotherhost]["success_ops"] == 1
        assert state.results[anotherhost]["ops"] == 1

        # And w/o errors
        assert state.results[somehost]["error_ops"] == 0
        assert state.results[anotherhost]["error_ops"] == 0

        # And with the different modes
        run_ops(state, serial=True)
        run_ops(state, no_wait=True)

        disconnect_all(state)

    # def test_op_call_direct_falls(self):
    #     inventory = make_inventory()
    #     somehost = inventory.get_host('somehost')
    #     state = State(inventory, Config())

    #     # Enable printing on this test to catch any exceptions in the formatting
    #     state.print_output = True
    #     state.print_input = True
    #     state.print_fact_info = True
    #     state.print_noop_info = True

    #     connect_all(state)

    #     with self.assertRaises(PyinfraError) as context:
    #         server.shell(commands='echo hi')

    #     assert context.exception.args[0] == (
    #         'Operation order number not provided in API mode - '
    #         'you must use `add_op` to add operations.'
    #     )

    @patch("pyinfra.api.util.open", mock_open(read_data="test!"), create=True)
    @patch("pyinfra.operations.files.os.path.isfile", lambda *args, **kwargs: True)
    def test_file_upload_op(self):
        inventory = make_inventory()

        state = State(inventory, Config())
        connect_all(state)

        # Test normal
        add_op(
            state,
            files.put,
            name="First op name",
            src="files/file.txt",
            dest="/home/vagrant/file.txt",
        )

        # And with sudo
        add_op(
            state,
            files.put,
            src="files/file.txt",
            dest="/home/vagrant/file.txt",
            sudo=True,
            sudo_user="pyinfra",
        )

        # And with su
        add_op(
            state,
            files.put,
            src="files/file.txt",
            dest="/home/vagrant/file.txt",
            sudo=True,
            su_user="pyinfra",
        )

        op_order = state.get_op_order()

        # Ensure we have all ops
        assert len(op_order) == 3

        first_op_hash = op_order[0]
        second_op_hash = op_order[1]

        # Ensure first op is the right one
        assert state.op_meta[first_op_hash]["names"] == {"First op name"}

        somehost = inventory.get_host("somehost")
        anotherhost = inventory.get_host("anotherhost")

        # Ensure first op has the right (upload) command
        assert state.ops[somehost][first_op_hash]["commands"] == [
            StringCommand("mkdir -p /home/vagrant"),
            FileUploadCommand("files/file.txt", "/home/vagrant/file.txt"),
        ]

        # Ensure second op has sudo/sudo_user
        assert state.ops[somehost][second_op_hash]["global_kwargs"]["sudo"] is True
        assert state.ops[somehost][second_op_hash]["global_kwargs"]["sudo_user"] == "pyinfra"

        # Ensure third has su_user
        assert state.ops[somehost][op_order[2]]["global_kwargs"]["su_user"] == "pyinfra"

        # Check run ops works
        run_ops(state)

        # Ensure ops completed OK
        assert state.results[somehost]["success_ops"] == 3
        assert state.results[somehost]["ops"] == 3
        assert state.results[anotherhost]["success_ops"] == 3
        assert state.results[anotherhost]["ops"] == 3

        # And w/o errors
        assert state.results[somehost]["error_ops"] == 0
        assert state.results[anotherhost]["error_ops"] == 0

    def test_file_download_op(self):
        inventory = make_inventory()

        state = State(inventory, Config())
        connect_all(state)

        with patch("pyinfra.operations.files.os.path.isfile", lambda *args, **kwargs: True):
            add_op(
                state,
                files.get,
                name="First op name",
                src="/home/vagrant/file.txt",
                dest="files/file.txt",
            )

        op_order = state.get_op_order()

        assert len(op_order) == 1

        first_op_hash = op_order[0]
        assert state.op_meta[first_op_hash]["names"] == {"First op name"}

        somehost = inventory.get_host("somehost")
        anotherhost = inventory.get_host("anotherhost")

        # Ensure first op has the right (upload) command
        assert state.ops[somehost][first_op_hash]["commands"] == [
            FileDownloadCommand("/home/vagrant/file.txt", "files/file.txt"),
        ]

        with patch("pyinfra.api.util.open", mock_open(read_data="test!"), create=True):
            run_ops(state)

        assert state.results[somehost]["success_ops"] == 1
        assert state.results[somehost]["ops"] == 1
        assert state.results[anotherhost]["success_ops"] == 1
        assert state.results[anotherhost]["ops"] == 1
        assert state.results[somehost]["error_ops"] == 0
        assert state.results[anotherhost]["error_ops"] == 0

    def test_function_call_op(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)

        is_called = []

        def mocked_function(*args, **kwargs):
            is_called.append(True)
            return None

        # Add op to both hosts
        add_op(state, python.call, mocked_function)

        # Ensure there is one op
        assert len(state.get_op_order()) == 1

        run_ops(state)

        assert is_called

    def test_run_once_serial_op(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)

        # Add a run once op
        add_op(state, server.shell, 'echo "hi"', run_once=True, serial=True)

        # Ensure it's added to op_order
        assert len(state.get_op_order()) == 1

        somehost = inventory.get_host("somehost")
        anotherhost = inventory.get_host("anotherhost")

        # Ensure between the two hosts we only run the one op
        assert len(state.ops[somehost]) + len(state.ops[anotherhost]) == 1

        # Check run works
        run_ops(state)

        assert (
            state.results[somehost]["success_ops"] + state.results[anotherhost]["success_ops"]
        ) == 1

    def test_rsync_op(self):
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        connect_all(state)

        with patch("pyinfra.connectors.ssh.check_can_rsync"):
            add_op(state, files.rsync, "src", "dest", sudo=True, sudo_user="root")

        assert len(state.get_op_order()) == 1

        with patch("pyinfra.connectors.ssh.run_local_process") as fake_run_local_process:
            fake_run_local_process.return_value = 0, []
            run_ops(state)

        fake_run_local_process.assert_called_with(
            (
                "rsync -ax --delete --rsh "
                "'ssh -o BatchMode=yes '"
                " --rsync-path 'sudo -u root rsync' src vagrant@somehost:dest"
            ),
            print_output=False,
            print_prefix=inventory.get_host("somehost").print_prefix,
        )

    def test_rsync_op_failure(self):
        inventory = make_inventory(hosts=("somehost",))
        state = State(inventory, Config())
        connect_all(state)

        with patch("pyinfra.connectors.ssh.find_executable", lambda x: None):
            with self.assertRaises(OperationError) as context:
                add_op(state, files.rsync, "src", "dest")

        assert context.exception.args[0] == "The `rsync` binary is not available on this system."

    def test_op_cannot_change_execution_kwargs(self):
        inventory = make_inventory()

        state = State(inventory, Config())

        class NoSetDefaultDict(defaultdict):
            def setdefault(self, key, _):
                return self[key]

        state.op_meta = NoSetDefaultDict(lambda: {"serial": True})

        connect_all(state)

        with self.assertRaises(OperationValueError) as context:
            add_op(state, files.file, "/var/log/pyinfra.log", serial=False)

        assert context.exception.args[0] == "Cannot have different values for `serial`."


class TestNestedOperationsApi(PatchSSHTestCase):
    def test_nested_op_api(self):
        inventory = make_inventory()
        state = State(inventory, Config())

        connect_all(state)

        somehost = inventory.get_host("somehost")

        ctx_state.set(state)
        ctx_host.set(somehost)

        pyinfra.is_cli = True

        try:
            outer_result = server.shell(commands="echo outer")
            assert outer_result.combined_output_lines is None

            def callback():
                inner_result = server.shell(commands="echo inner")
                assert inner_result.combined_output_lines is not None

            python.call(function=callback)

            assert len(state.get_op_order()) == 2

            run_ops(state)

            assert len(state.get_op_order()) == 3
            assert state.results[somehost]["success_ops"] == 3
            assert outer_result.combined_output_lines is not None

            disconnect_all(state)
        finally:
            pyinfra.is_cli = False


class TestOperationFailures(PatchSSHTestCase):
    def test_full_op_fail(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)

        add_op(state, server.shell, 'echo "hi"')

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_channel = FakeChannel(1)
            fake_run_command.return_value = (
                False,
                FakeBuffer("", fake_channel),
            )

            with self.assertRaises(PyinfraError) as e:
                run_ops(state)

            assert e.exception.args[0] == "No hosts remaining!"

            somehost = inventory.get_host("somehost")

            # Ensure the op was not flagged as success
            assert state.results[somehost]["success_ops"] == 0
            # And was flagged asn an error
            assert state.results[somehost]["error_ops"] == 1

    def test_ignore_errors_op_fail(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)

        add_op(state, server.shell, 'echo "hi"', ignore_errors=True)

        with patch("pyinfra.connectors.ssh.run_shell_command") as fake_run_command:
            fake_channel = FakeChannel(1)
            fake_run_command.return_value = (
                False,
                FakeBuffer("", fake_channel),
            )

            # This should run OK
            run_ops(state)

        somehost = inventory.get_host("somehost")

        # Ensure the op was added to results
        assert state.results[somehost]["ops"] == 1
        assert state.results[somehost]["ignored_error_ops"] == 1
        # But not as a success
        assert state.results[somehost]["success_ops"] == 0

    # def test_no_invalid_op_call(self):
    #     inventory = make_inventory()
    #     state = State(inventory, Config())
    #     connect_all(state)
    #     ctx_state.set(state)

    #     state.in_op = True
    #     with self.assertRaises(PyinfraError):
    #         server.user('user')

    #     state.in_op = False
    #     state.in_deploy = True
    #     with self.assertRaises(PyinfraError):
    #         server.user('user')


class TestOperationOrdering(PatchSSHTestCase):
    # In CLI mode, pyinfra uses *line numbers* to order operations as defined by
    # the user. This makes reasoning about user-written deploys simple and easy
    # to understand.
    def test_cli_op_line_numbers(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)

        state.current_deploy_filename = __file__

        pyinfra.is_cli = True
        ctx_state.set(state)

        # Add op to both hosts
        for name in ("anotherhost", "somehost"):
            ctx_host.set(inventory.get_host(name))
            server.shell("echo hi")  # note this is called twice but on *the same line*

        # Add op to just the second host - using the context modules such that
        # it replicates a deploy file.
        ctx_host.set(inventory.get_host("anotherhost"))
        first_context_hash = server.user("anotherhost_user").hash

        # Add op to just the first host - using the context modules such that
        # it replicates a deploy file.
        ctx_host.set(inventory.get_host("somehost"))
        second_context_hash = server.user("somehost_user").hash

        ctx_state.reset()
        ctx_host.reset()

        pyinfra.is_cli = False

        # Ensure there are two ops
        op_order = state.get_op_order()
        assert len(op_order) == 3

        # And that the two ops above were called in the expected order
        assert op_order[1] == first_context_hash
        assert op_order[2] == second_context_hash

        # Ensure somehost has two ops and anotherhost only has the one
        assert len(state.ops[inventory.get_host("somehost")]) == 2
        assert len(state.ops[inventory.get_host("anotherhost")]) == 2

    # In API mode, pyinfra *overrides* the line numbers such that whenever an
    # operation or deploy is added it is simply appended. This makes sense as
    # the user writing the API calls has full control over execution order.
    def test_api_op_line_numbers(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)

        another_host = inventory.get_host("anotherhost")

        def add_another_op():
            return add_op(state, server.shell, "echo second-op")[another_host].hash

        first_op_hash = add_op(state, server.shell, "echo first-op")[another_host].hash
        second_op_hash = add_another_op()  # note `add_op` will be called on an earlier line

        op_order = state.get_op_order()
        assert len(op_order) == 2

        assert op_order[0] == first_op_hash
        assert op_order[1] == second_op_hash


this_filename = path.join("tests", "test_api", "test_api_operations.py")


# class TestOperationExceptions(TestCase):
#     def test_add_op_rejects_cli(self):
#         pyinfra.is_cli = True

#         with self.assertRaises(PyinfraError) as context:
#             add_op(None, server.shell)
#         call_line = getframeinfo(currentframe()).lineno - 1

#         pyinfra.is_cli = False

#         assert context.exception.args[0] == (
#             '`add_op` should not be called when pyinfra is executing in CLI mode! '
#             '(line {0} in {1})'.format(call_line, this_filename)
#         )

#     def test_op_call_rejects_no_cli(self):
#         with self.assertRaises(PyinfraError) as context:
#             server.shell()
#         call_line = getframeinfo(currentframe()).lineno - 1

#         assert context.exception.args[0] == (
#             'API operation called without state/host: '
#             'server.shell (line {0} in {1})'.format(call_line, this_filename)
#         )

#     def test_op_call_rejects_in_op(self):
#         state = FakeState()

#         pyinfra.is_cli = True
#         ctx_state.set(state)

#         with self.assertRaises(PyinfraError) as context:
#             server.shell()
#         call_line = getframeinfo(currentframe()).lineno - 1

#         pyinfra.is_cli = False
#         ctx_state.reset()

#         assert context.exception.args[0] == (
#             'Nested operation called without state/host: '
#             'server.shell (line {0} in {1})'.format(call_line, this_filename)
#         )

#     def test_op_call_rejects_in_deploy(self):
#         state = FakeState()
#         state.in_op = False

#         pyinfra.is_cli = True
#         ctx_state.set(state)

#         with self.assertRaises(PyinfraError) as context:
#             server.shell()
#         call_line = getframeinfo(currentframe()).lineno - 1

#         pyinfra.is_cli = False
#         ctx_state.reset()

#         assert context.exception.args[0] == (
#             'Nested deploy operation called without state/host: '
#             'server.shell (line {0} in {1})'.format(call_line, this_filename)
#         )



import json

from pyinfra.api import FactBase


class OpenvzContainers(FactBase):
    """
    Returns a dict of running OpenVZ containers by CTID:

    .. code:: python

        {
            666: {
                "ip": [],
                "ostemplate": "ubuntu...",
                ...
            },
        }
    """

    command = "vzlist -a -j"
    requires_command = "vzlist"

    default = dict

    @staticmethod
    def process(output):
        combined_json = "".join(output)
        vz_data = json.loads(combined_json)

        return {int(vz.pop("ctid")): vz for vz in vz_data}



"""
The windows_files module handles windows filesystem state, file uploads and template generation.
"""

import ntpath
import os
from datetime import timedelta

from pyinfra import host, state
from pyinfra.api import FileUploadCommand, OperationError, OperationTypeError, operation
from pyinfra.api.util import get_file_sha1
from pyinfra.facts.windows import Date
from pyinfra.facts.windows_files import Directory, File, Link, Md5File, Sha1File, Sha256File

from .util.files import ensure_mode_int


@operation(
    pipeline_facts={"file": "dest"},
)
def download(
    src,
    dest,
    user=None,
    group=None,
    mode=None,
    cache_time=None,
    force=False,
    sha256sum=None,
    sha1sum=None,
    md5sum=None,
):
    """
    Download files from remote locations using curl or wget.

    + src: source URL of the file
    + dest: where to save the file
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files
    + cache_time: if the file exists already, re-download after this time (in seconds)
    + force: always download the file, even if it already exists
    + sha256sum: sha256 hash to checksum the downloaded file against
    + sha1sum: sha1 hash to checksum the downloaded file against
    + md5sum: md5 hash to checksum the downloaded file against

    **Example:**

    .. code:: python

        winows_files.download(
            name="Download the Docker repo file",
            src="https://download.docker.com/linux/centos/docker-ce.repo",
            dest="C:\\docker",
        )
    """

    info = host.get_fact(File, path=dest)
    # Destination is a directory?
    if info is False:
        raise OperationError(
            "Destination {0} already exists and is not a file".format(dest),
        )

    # Do we download the file? Force by default
    download = force

    # Doesn't exist, lets download it
    if info is None:
        download = True

    # Destination file exists & cache_time: check when the file was last modified,
    # download if old
    else:
        if cache_time:
            # Time on files is not tz-aware, and will be the same tz as the server's time,
            # so we can safely remove the tzinfo from Date before comparison.
            cache_time = host.get_fact(Date).replace(tzinfo=None) - timedelta(seconds=cache_time)
            if info["mtime"] and info["mtime"] > cache_time:
                download = True

        if sha1sum:
            if sha1sum != host.get_fact(Sha1File, path=dest):
                download = True

        if sha256sum:
            if sha256sum != host.get_fact(Sha256File, path=dest):
                download = True

        if md5sum:
            if md5sum != host.get_fact(Md5File, path=dest):
                download = True

    # If we download, always do user/group/mode as SSH user may be different
    if download:
        yield (
            '$ProgressPreference = "SilentlyContinue"; ' "Invoke-WebRequest -Uri {0} -OutFile {1}"
        ).format(src, dest)

        # if user or group:
        #    yield chown(dest, user, group)

        # if mode:
        #    yield chmod(dest, mode)

        if sha1sum:
            yield (
                'if ((Get-FileHash -Algorithm SHA1 "{0}").hash -ne {1}) {{ '
                'Write-Error "SHA1 did not match!" '
                "}}"
            ).format(dest, sha1sum)

        if sha256sum:
            yield (
                'if ((Get-FileHash -Algorithm SHA256 "{0}").hash -ne {1}) {{ '
                'Write-Error "SHA256 did not match!" '
                "}}"
            ).format(dest, sha256sum)

        if md5sum:
            yield (
                'if ((Get-FileHash -Algorithm MD5 "{0}").hash -ne {1}) {{ '
                'Write-Error "MD5 did not match!" '
                "}}"
            ).format(dest, md5sum)

    else:
        host.noop("file {0} has already been downloaded".format(dest))


@operation(
    pipeline_facts={
        "file": "dest",
        "sha1_file": "dest",
    },
)
def put(
    src,
    dest,
    user=None,
    group=None,
    mode=None,
    add_deploy_dir=True,
    create_remote_dir=True,
    force=False,
    assume_exists=False,
):
    """
    Upload a local file to the remote system.

    + src: local filename to upload
    + dest: remote filename to upload to
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files
    + add_deploy_dir: src is relative to the deploy directory
    + create_remote_dir: create the remote directory if it doesn't exist
    + force: always upload the file, even if the remote copy matches
    + assume_exists: whether to assume the local file exists

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Note:
        This operation is not suitable for large files as it may involve copying
        the file before uploading it.

    **Examples:**

    .. code:: python

        # Note: This requires a 'files/motd' file on the local filesystem
        files.put(
            name="Update the message of the day file",
            src="data/content.json",
            dest="C:\\data\\content.json",
        )
    """

    # Upload IO objects as-is
    if hasattr(src, "read"):
        local_file = src

    # Assume string filename
    else:
        # Add deploy directory?
        if add_deploy_dir and state.cwd:
            src = os.path.join(state.cwd, src)

        local_file = src

        if not assume_exists and not os.path.isfile(local_file):
            raise IOError("No such file: {0}".format(local_file))

    mode = ensure_mode_int(mode)
    remote_file = host.get_fact(File, path=dest)

    if create_remote_dir:
        yield from _create_remote_dir(state, host, dest, user, group)

    # No remote file, always upload and user/group/mode if supplied
    if not remote_file or force:
        yield FileUploadCommand(
            local_file,
            dest,
            remote_temp_filename=state.get_temp_filename(dest),
        )

        # if user or group:
        #    yield chown(dest, user, group)

        # if mode:
        #    yield chmod(dest, mode)

    # File exists, check sum and check user/group/mode if supplied
    else:
        local_sum = get_file_sha1(src)
        remote_sum = host.get_fact(Sha1File, path=dest)

        # Check sha1sum, upload if needed
        if local_sum != remote_sum:
            yield FileUploadCommand(
                local_file,
                dest,
                remote_temp_filename=state.get_temp_filename(dest),
            )

            # if user or group:
            #    yield chown(dest, user, group)

            # if mode:
            #    yield chmod(dest, mode)

        else:
            changed = False

            # Check mode
            # if mode and remote_file['mode'] != mode:
            #    yield chmod(dest, mode)
            #    changed = True

            # Check user/group
            # if (
            #    (user and remote_file['user'] != user)
            #    or (group and remote_file['group'] != group)
            # ):
            #    yield chown(dest, user, group)
            #    changed = True

            if not changed:
                host.noop("file {0} is already uploaded".format(dest))


@operation(
    pipeline_facts={"windows_file": "name"},
)
def file(
    path,
    present=True,
    assume_present=False,
    user=None,
    group=None,
    mode=None,
    touch=False,
    create_remote_dir=True,
):
    """
    Add/remove/update files.

    + path: path of the remote file
    + present: whether the file should exist
    + assume_present: whether to assume the file exists
    + TODO: user: user to own the files
    + TODO: group: group to own the files
    + TODO: mode: permissions of the files as an integer, eg: 755
    + touch: whether to touch the file
    + create_remote_dir: create the remote directory if it doesn't exist

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    **Example:**

    .. code:: python

        files.file(
            name="Create c:\\temp\\hello.txt",
            path="c:\\temp\\hello.txt",
            touch=True,
        )
    """

    if not isinstance(path, str):
        raise OperationTypeError("Name must be a string")

    # mode = ensure_mode_int(mode)
    info = host.get_fact(File, path=path)

    # Not a file?!
    if info is False:
        raise OperationError("{0} exists and is not a file".format(path))

    # Doesn't exist & we want it
    if not assume_present and info is None and present:
        if create_remote_dir:
            yield from _create_remote_dir(state, host, path, user, group)

        yield "New-Item -ItemType file {0}".format(path)

    #        if mode:
    #            yield chmod(path, mode)
    #        if user or group:
    #            yield chown(path, user, group)

    # It exists and we don't want it
    elif (assume_present or info) and not present:
        yield "Remove-Item {0}".format(path)


#    # It exists & we want to ensure its state
#    elif (assume_present or info) and present:
#        if touch:
#            yield 'New-Item -ItemType file {0}'.format(path)
#
#        # Check mode
#        if mode and (not info or info['mode'] != mode):
#            yield chmod(path, mode)
#
#        # Check user/group
#        if (
#            (not info and (user or group))
#            or (user and info['user'] != user)
#            or (group and info['group'] != group)
#        ):
#            yield chown(path, user, group)


def _create_remote_dir(state, host, remote_filename, user, group):
    # Always use POSIX style path as local might be Windows, remote always *nix
    remote_dirname = ntpath.dirname(remote_filename)
    if remote_dirname:
        yield from directory(
            remote_dirname,
            state=state,
            host=host,
            user=user,
            group=group,
        )


@operation(
    pipeline_facts={"windows_directory": "name"},
)
def directory(
    path,
    present=True,
    assume_present=False,
    user=None,
    group=None,
    mode=None,
    recursive=False,
):
    """
    Add/remove/update directories.

    + path: path of the remote folder
    + present: whether the folder should exist
    + assume_present: whether to assume the directory exists
    + TODO: user: user to own the folder
    + TODO: group: group to own the folder
    + TODO: mode: permissions of the folder
    + TODO: recursive: recursively apply user/group/mode

    **Examples:**

    .. code:: python

        files.directory(
            name="Ensure the c:\\temp\\dir_that_we_want_removed is removed",
            path="c:\\temp\\dir_that_we_want_removed",
            present=False,
        )

        files.directory(
            name="Ensure c:\\temp\\foo\\foo_dir exists",
            path="c:\\temp\\foo\\foo_dir",
            recursive=True,
        )

        # multiple directories
        dirs = ["c:\\temp\\foo_dir1", "c:\\temp\\foo_dir2"]
        for dir in dirs:
            files.directory(
                name="Ensure the directory `{}` exists".format(dir),
                path=dir,
            )

    """

    if not isinstance(path, str):
        raise OperationTypeError("Name must be a string")

    info = host.get_fact(Directory, path=path)

    # Not a directory?!
    if info is False:
        raise OperationError("{0} exists and is not a directory".format(path))

    # Doesn't exist & we want it
    if not assume_present and info is None and present:
        yield "New-Item -Path {0} -ItemType Directory".format(path)
        #        if mode:
        #            yield chmod(path, mode, recursive=recursive)
        #        if user or group:
        #            yield chown(path, user, group, recursive=recursive)
        #
        # Somewhat bare fact, should flesh out more
        host.create_fact(
            Date,
            kwargs={"path": path},
            data={"type": "directory"},
        )

    # It exists and we don't want it
    elif (assume_present or info) and not present:
        # TODO: how to ensure we use 'ps'?
        # remove anything in the directory
        yield "Get-ChildItem {0} -Recurse | Remove-Item".format(path)
        # remove directory
        yield "Remove-Item {0}".format(path)

    # It exists & we want to ensure its state


#    elif (assume_present or info) and present:
#        # Check mode
#        if mode and (not info or info['mode'] != mode):
#            yield chmod(path, mode, recursive=recursive)
#
#        # Check user/group
#        if (
#            (not info and (user or group))
#            or (user and info['user'] != user)
#            or (group and info['group'] != group)
#        ):
#            yield chown(path, user, group, recursive=recursive)


def _validate_path(path):
    try:
        path = os.fspath(path)
    except TypeError:
        raise OperationTypeError("`path` must be a string or `os.PathLike` object")


@operation(
    pipeline_facts={"link": "path"},
)
def link(
    path,
    target=None,
    present=True,
    assume_present=False,
    user=None,
    group=None,
    symbolic=True,
    force=True,
    create_remote_dir=True,
):
    """
    Add/remove/update links.

    + path: the name of the link
    + target: the file/directory the link points to
    + present: whether the link should exist
    + assume_present: whether to assume the link exists
    + user: user to own the link
    + group: group to own the link
    + symbolic: whether to make a symbolic link (vs hard link)
    + create_remote_dir: create the remote directory if it doesn't exist

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Source changes:
        If the link exists and points to a different target, pyinfra will remove it and
        recreate a new one pointing to then new target.

    **Examples:**

    .. code:: python

        # simple example showing how to link to a file
        files.link(
            name=r"Create link C:\\issue2 that points to C:\\issue",
            path=r"C:\\issue2",
            target=r"C\\issue",
        )
    """

    _validate_path(path)

    if present and not target:
        raise OperationError("If present is True target must be provided")

    info = host.get_fact(Link, path=path)

    # Not a link?
    if info is not None and not info:
        raise OperationError("{0} exists and is not a link".format(path))

    add_cmd = "New-Item -ItemType {0} -Path {1} -Target {2} {3}".format(
        "SymbolicLink" if symbolic else "HardLink",
        path,
        target,
        "-Force" if force else "",
    )

    remove_cmd = "(Get-Item {0}).Delete()".format(path)

    # We will attempt to link regardless of current existence
    # since we know by now the path is either a link already
    # or does not exist
    if (info is None or force) and present:
        if create_remote_dir:
            yield from _create_remote_dir(state, host, path, user, group)

        yield add_cmd

        # if user or group:
        #    yield chown(path, user, group, dereference=False)

        # host.create_fact(
        #    WindowsLink,
        #    kwargs={'name': path},
        #    data={'link_target': target, 'group': group, 'user': user},
        # )

    # It exists and we don't want it
    elif (assume_present or info) and not present:
        yield remove_cmd
        # host.delete_fact(WindowsLink, kwargs={'name': path})

    else:
        host.noop("link {0} already exists and force=False".format(path))



import pkg_resources


class BaseConnectorMeta:
    handles_execution = False
    keys_prefix = ""

    class DataKeys:
        pass

    @classmethod
    def keys(cls):
        class Keys:
            pass

        for key in cls.DataKeys.__dict__:
            if not key.startswith("_"):
                setattr(Keys, key, f"{cls.keys_prefix}_{key}")

        return Keys


def _load_connector(entrypoint):
    connector = entrypoint.load()
    if not getattr(connector, "Meta", None):
        connector.Meta = BaseConnectorMeta
    return connector


def get_all_connectors():
    return {
        entrypoint.name: _load_connector(entrypoint)
        for entrypoint in pkg_resources.iter_entry_points("pyinfra.connectors")
    }


def get_execution_connectors():
    return {
        connector: connector_mod
        for connector, connector_mod in get_all_connectors().items()
        if connector_mod.Meta.handles_execution
    }


def get_execution_connector(name):
    return get_execution_connectors()[name]



"""
The ``@local`` connector executes changes on the local machine using subprocesses.
"""

import os
from distutils.spawn import find_executable
from tempfile import mkstemp
from typing import TYPE_CHECKING

import click

from pyinfra import logger
from pyinfra.api.connectors import BaseConnectorMeta
from pyinfra.api.exceptions import InventoryError
from pyinfra.api.util import get_file_io

from .util import (
    execute_command_with_sudo_retry,
    make_unix_command_for_host,
    run_local_process,
    split_combined_output,
)

if TYPE_CHECKING:
    from pyinfra.api.host import Host
    from pyinfra.api.state import State


class Meta(BaseConnectorMeta):
    handles_execution = True


def make_names_data(_=None):
    if _ is not None:
        raise InventoryError("Cannot have more than one @local")

    yield "@local", {}, ["@local"]


def connect(state: "State", host: "Host"):
    return True


def run_shell_command(
    state: "State",
    host: "Host",
    command,
    get_pty: bool = False,  # ignored
    timeout=None,
    stdin=None,
    success_exit_codes=None,
    print_output: bool = False,
    print_input: bool = False,
    return_combined_output: bool = False,
    **command_kwargs,
):
    """
    Execute a command on the local machine.

    Args:
        state (``pyinfra.api.State`` object): state object for this command
        host (``pyinfra.api.Host`` object): the target host
        command (string): actual command to execute
        sudo (boolean): whether to wrap the command with sudo
        sudo_user (string): user to sudo to
        env (dict): environment variables to set
        timeout (int): timeout for this command to complete before erroring

    Returns:
        tuple: (exit_code, stdout, stderr)
        stdout and stderr are both lists of strings from each buffer.
    """

    def execute_command():
        unix_command = make_unix_command_for_host(state, host, command, **command_kwargs)
        actual_command = unix_command.get_raw_value()

        logger.debug("--> Running command on localhost: %s", unix_command)

        if print_input:
            click.echo("{0}>>> {1}".format(host.print_prefix, unix_command), err=True)

        return run_local_process(
            actual_command,
            stdin=stdin,
            timeout=timeout,
            print_output=print_output,
            print_prefix=host.print_prefix,
        )

    return_code, combined_output = execute_command_with_sudo_retry(
        host,
        command_kwargs,
        execute_command,
    )

    if success_exit_codes:
        status = return_code in success_exit_codes
    else:
        status = return_code == 0

    if return_combined_output:
        return status, combined_output

    stdout, stderr = split_combined_output(combined_output)
    return status, stdout, stderr


def put_file(
    state: "State",
    host: "Host",
    filename_or_io,
    remote_filename,
    remote_temp_filename=None,  # ignored
    print_output: bool = False,
    print_input: bool = False,
    **command_kwargs,
):
    """
    Upload a local file or IO object by copying it to a temporary directory
    and then writing it to the upload location.
    """

    _, temp_filename = mkstemp()

    try:
        # Load our file or IO object and write it to the temporary file
        with get_file_io(filename_or_io) as file_io:
            with open(temp_filename, "wb") as temp_f:
                data = file_io.read()

                if isinstance(data, str):
                    data = data.encode()

                temp_f.write(data)

        # Copy the file using `cp` such that we support sudo/su
        status, _, stderr = run_shell_command(
            state,
            host,
            "cp {0} {1}".format(temp_filename, remote_filename),
            print_output=print_output,
            print_input=print_input,
            **command_kwargs,
        )

        if not status:
            raise IOError("\n".join(stderr))
    finally:
        os.remove(temp_filename)

    if print_output:
        click.echo(
            "{0}file copied: {1}".format(host.print_prefix, remote_filename),
            err=True,
        )

    return status


def get_file(
    state: "State",
    host: "Host",
    remote_filename,
    filename_or_io,
    remote_temp_filename=None,  # ignored
    print_output: bool = False,
    print_input: bool = False,
    **command_kwargs,
):
    """
    Download a local file by copying it to a temporary location and then writing
    it to our filename or IO object.
    """

    _, temp_filename = mkstemp()

    try:
        # Copy the file using `cp` such that we support sudo/su
        status, _, stderr = run_shell_command(
            state,
            host,
            "cp {0} {1}".format(remote_filename, temp_filename),
            print_output=print_output,
            print_input=print_input,
            **command_kwargs,
        )

        if not status:
            raise IOError("\n".join(stderr))

        # Load our file or IO object and write it to the temporary file
        with open(temp_filename) as temp_f:
            with get_file_io(filename_or_io, "wb") as file_io:
                data_bytes: bytes

                data = temp_f.read()
                if isinstance(data, str):
                    data_bytes = data.encode()
                else:
                    data_bytes = data

                file_io.write(data_bytes)
    finally:
        os.remove(temp_filename)

    if print_output:
        click.echo(
            "{0}file copied: {1}".format(host.print_prefix, remote_filename),
            err=True,
        )

    return True


def check_can_rsync(host):
    if not find_executable("rsync"):
        raise NotImplementedError("The `rsync` binary is not available on this system.")


def rsync(
    state: "State",
    host: "Host",
    src,
    dest,
    flags,
    print_output: bool = False,
    print_input: bool = False,
    **command_kwargs,
):
    status, _, stderr = run_shell_command(
        state,
        host,
        "rsync {0} {1} {2}".format(" ".join(flags), src, dest),
        print_output=print_output,
        print_input=print_input,
        **command_kwargs,
    )

    if not status:
        raise IOError("\n".join(stderr))

    return True



import re
from datetime import datetime

from dateutil.parser import parse as parse_date

from pyinfra.api import FactBase


class Home(FactBase):
    """
    Returns the home directory of the current user.
    """

    command = "echo %HOMEPATH%"
    shell_executable = "cmd"

    @staticmethod
    def process(output):
        return "".join(output).replace("\n", "")


class Hostname(FactBase):
    """
    Returns the current hostname of the server.
    """

    command = "hostname"

    @staticmethod
    def process(output):
        return "".join(output).replace("\n", "")


class LastReboot(FactBase):
    """
    Returns the date and time of the last reboot.
    """

    command = (
        "Get-CimInstance -ClassName Win32_OperatingSystem | "
        "Select -ExpandProperty LastBootUptime"
    )

    @staticmethod
    def process(output):
        return "".join(output).replace("\n", "")


class Os(FactBase):
    """
    Returns the OS name according to ``systeminfo``.
    """

    command = 'systeminfo.exe | findstr /c:"OS Name:"'

    @staticmethod
    def process(output):
        new_output = ""
        match = re.match("OS Name:[ ]*(.*)", output[0])
        if match:
            new_output = match.group(1)
        return new_output


class Bios(FactBase):
    """
    Returns the BIOS info.
    """

    command = "Get-CimInstance -ClassName Win32_BIOS"

    @staticmethod
    def process(output):
        bios = {}
        for line in output:
            line_data = line.split(":")
            if len(line_data) > 1:
                bios.update({line_data[0].strip(): line_data[1].strip()})
        return bios


def _format_windows(output):
    lines = {}
    for line in output:
        # split line on ':'
        line_data = line.split(":")
        if len(line_data) > 1:
            # we have a data line
            this_key = line_data[0].strip()
            this_data = line_data[1].strip()
            if len(line_data) > 2:
                # there was a ':' in the data, so reconstitute the value
                this_data = ":".join(line_data[1:]).strip()
            lines[this_key] = this_data
    return lines


def _format_windows_for_key(primary_key, output, return_primary_key=True):
    """Format the windows powershell output that uses 'Format-Line'
    into a dict of dicts.
    """
    primary_key = primary_key.strip()
    lines = {}
    one_item = {}
    key_value = ""
    for line in output:
        # split line on ':'
        line_data = line.split(":")
        if len(line_data) > 1:
            # we have a data line
            this_key = line_data[0].strip()
            this_data = line_data[1].strip()
            if len(line_data) > 2:
                # there was a ':' in the data, so reconstitute the value
                this_data = ":".join(line_data[1:]).strip()
            if this_key != primary_key:
                one_item.update({this_key: this_data})
            else:
                key_value = this_data
        else:
            if line == "":
                if one_item:
                    lines[key_value] = one_item
                    one_item = {}
                    key_value = ""
            else:
                # append this_data to the existing entry
                this_data = line.strip()
                appended_data = one_item[this_key] + this_data
                one_item.update({this_key: appended_data})
    if return_primary_key:
        return {primary_key: lines}
    else:
        return lines


class Processors(FactBase):
    """
    Returns the processors info.
    """

    command = "Get-CimInstance -ClassName Win32_Processor | Format-List -Property *"

    @staticmethod
    def process(output):
        return _format_windows_for_key("DeviceID", output)


class OsVersion(FactBase):
    """
    Returns the OS version according to ``systeminfo``.
    """

    command = 'systeminfo | findstr /c:"OS Version:"'

    @staticmethod
    def process(output):
        new_output = ""
        match = re.match("OS Version:[ ]*(.*)", output[0])
        if match:
            new_output = match.group(1)
        return new_output


class SystemType(FactBase):
    """
    Returns the system type according to ``systeminfo``.
    """

    command = 'systeminfo | findstr /c:"System Type:"'

    @staticmethod
    def process(output):
        new_output = ""
        match = re.match("System Type:[ ]*(.*)", output[0])
        if match:
            new_output = match.group(1)
        return new_output


class Date(FactBase):
    """
    Returns the current datetime on the server.
    """

    command = "echo %date%-%time%"
    shell_executable = "cmd"
    default = datetime.now

    @staticmethod
    def process(output):
        new_output = "".join(output)
        return parse_date(new_output)


class LocalGroups(FactBase):
    """
    Returns a list of groups on the system.
    """

    command = "net localgroup | findstr [^*]"

    default = list

    @staticmethod
    def process(output):
        groups = []
        for group in output:
            # Note: If run this command thru ps, there are headers/footer.
            # remove empty groups and those groups that are not local
            if group != "" and group[0] == "*":
                groups.append(group)
        return groups


class Where(FactBase):
    """
    Returns the full path for a command, if available.
    """

    shell_executable = "cmd"

    @staticmethod
    def command(name):
        return "where {0}".format(name)

    @staticmethod
    def process(output):
        return output[0].rstrip()


class Hotfixes(FactBase):
    """
    Returns the Windows hotfixes.
    """

    command = "Get-CimInstance -ClassName Win32_QuickFixEngineering | Format-List -Property *"

    @staticmethod
    def process(output):
        return _format_windows_for_key("HotFixID", output)


class LocalDrivesInfo(FactBase):
    """
    Returns the Windows local drives info.
    """

    command = (
        'Get-CimInstance -ClassName Win32_LogicalDisk -Filter "DriveType=3" '
        "| Format-List -Property *"
    )

    @staticmethod
    def process(output):
        return _format_windows_for_key("DeviceID", output)


class LoggedInUserInfo(FactBase):
    """
    Returns the Windows user logged in info.
    """

    command = (
        "Get-CimInstance -ClassName Win32_ComputerSystem -Property UserName "
        "| Format-List -Property *"
    )

    @staticmethod
    def process(output):
        return _format_windows_for_key("Name", output)


class LogonSessionInfo(FactBase):
    """
    Returns the Windows user logon session info.
    """

    command = "Get-CimInstance -ClassName Win32_LogonSession | Format-List -Property *"

    @staticmethod
    def process(output):
        return _format_windows_for_key("LogonId", output)


class Aliases(FactBase):
    """
    Returns the Windows aliases.
    """

    command = "Get-Alias | Format-List -Property *"

    @staticmethod
    def process(output):
        return _format_windows_for_key("Name", output)


class Services(FactBase):
    """
    Returns the Windows services.
    """

    command = "Get-CimInstance -ClassName Win32_Service | Format-List -Property *"

    @staticmethod
    def process(output):
        return _format_windows_for_key("Name", output)


class Service(FactBase):
    """
    Returns info about a Windows service.
    """

    def command(self, name):
        return "Get-Service -Name {} | Format-List -Property *".format(name)

    def process(self, output):
        return _format_windows_for_key("Name", output, return_primary_key=False)


class Processes(FactBase):
    """
    Returns the Windows processes.
    """

    command = "Get-Process | Format-List -Property *"

    @staticmethod
    def process(output):
        return _format_windows_for_key("Id", output)


class NetworkConfiguration(FactBase):
    """
    Returns the Windows network configuration.
    """

    command = "Get-CimInstance -Class Win32_NetworkAdapterConfiguration | Format-List -Property *"

    @staticmethod
    def process(output):
        return _format_windows_for_key("Index", output)


class InstallerApplications(FactBase):
    """
    Returns the Windows installer applications.
    """

    command = "Get-CimInstance -Class Win32_Product | Format-List -Property *"

    @staticmethod
    def process(output):
        return _format_windows_for_key("IdentifyingNumber", output)


class ComputerInfo(FactBase):
    """
    Returns the Windows info.
    """

    command = "Get-ComputerInfo | Format-List -Property *"

    @staticmethod
    def process(output):
        return _format_windows(output)



import shlex
from getpass import getpass
from socket import timeout as timeout_error
from subprocess import PIPE, Popen

import click
import gevent
from gevent.queue import Queue

from pyinfra import logger
from pyinfra.api import MaskString, QuoteString, StringCommand
from pyinfra.api.util import memoize

SUDO_ASKPASS_ENV_VAR = "PYINFRA_SUDO_PASSWORD"
SUDO_ASKPASS_COMMAND = r"""
temp=$(mktemp /tmp/pyinfra-sudo-askpass-XXXXXXXXXXXX)
cat >"$temp"<<'__EOF__'
#!/bin/sh
printf '%s\n' "${0}"
__EOF__
chmod 755 "$temp"
echo "$temp"
""".format(
    SUDO_ASKPASS_ENV_VAR,
)


def read_buffer(type_, io, output_queue, print_output=False, print_func=None):
    """
    Reads a file-like buffer object into lines and optionally prints the output.
    """

    def _print(line):
        if print_func:
            line = print_func(line)

        click.echo(line, err=True)

    for line in io:
        # Handle local Popen shells returning list of bytes, not strings
        if not isinstance(line, str):
            line = line.decode("utf-8")

        line = line.rstrip("\n")
        output_queue.put((type_, line))

        if print_output:
            _print(line)


def execute_command_with_sudo_retry(host, command_kwargs, execute_command):
    return_code, combined_output = execute_command()

    if return_code != 0 and combined_output:
        last_line = combined_output[-1][1]
        if last_line.strip() == "sudo: a password is required":
            command_kwargs["use_sudo_password"] = True  # ask for the password
            return_code, combined_output = execute_command()

    return return_code, combined_output


def run_local_process(
    command,
    stdin=None,
    timeout=None,
    print_output=False,
    print_prefix=None,
):
    process = Popen(command, shell=True, stdout=PIPE, stderr=PIPE, stdin=PIPE)

    if stdin:
        write_stdin(stdin, process.stdin)

    combined_output = read_buffers_into_queue(
        process.stdout,
        process.stderr,
        timeout=timeout,
        print_output=print_output,
        print_prefix=print_prefix,
    )

    logger.debug("--> Waiting for exit status...")
    process.wait()
    logger.debug("--> Command exit status: %i", process.returncode)

    # Close any open file descriptors
    process.stdout.close()
    process.stderr.close()

    return process.returncode, combined_output


def read_buffers_into_queue(
    stdout_buffer,
    stderr_buffer,
    timeout,
    print_output,
    print_prefix,
):
    output_queue = Queue()

    # Iterate through outputs to get an exit status and generate desired list
    # output, done in two greenlets so stdout isn't printed before stderr. Not
    # attached to state.pool to avoid blocking it with 2x n-hosts greenlets.
    stdout_reader = gevent.spawn(
        read_buffer,
        "stdout",
        stdout_buffer,
        output_queue,
        print_output=print_output,
        print_func=lambda line: "{0}{1}".format(print_prefix, line),
    )
    stderr_reader = gevent.spawn(
        read_buffer,
        "stderr",
        stderr_buffer,
        output_queue,
        print_output=print_output,
        print_func=lambda line: "{0}{1}".format(
            print_prefix,
            click.style(line, "red"),
        ),
    )

    # Wait on output, with our timeout (or None)
    greenlets = gevent.wait((stdout_reader, stderr_reader), timeout=timeout)

    # Timeout doesn't raise an exception, but gevent.wait returns the greenlets
    # which did complete. So if both haven't completed, we kill them and fail
    # with a timeout.
    if len(greenlets) != 2:
        stdout_reader.kill()
        stderr_reader.kill()

        raise timeout_error()

    return list(output_queue.queue)


def split_combined_output(combined_output):
    stdout = []
    stderr = []

    for type_, line in combined_output:
        if type_ == "stdout":
            stdout.append(line)
        elif type_ == "stderr":
            stderr.append(line)
        else:  # pragma: no cover
            raise ValueError("Incorrect output line type: {0}".format(type_))

    return stdout, stderr


def write_stdin(stdin, buffer):
    if hasattr(stdin, "readlines"):
        stdin = stdin.readlines()
    if not isinstance(stdin, (list, tuple)):
        stdin = [stdin]

    for line in stdin:
        if not line.endswith("\n"):
            line = "{0}\n".format(line)
        line = line.encode()
        buffer.write(line)
    buffer.close()


def _get_sudo_password(host, use_sudo_password):
    if not host.connector_data.get("sudo_askpass_path"):
        _, stdout, _ = host.run_shell_command(SUDO_ASKPASS_COMMAND)
        host.connector_data["sudo_askpass_path"] = shlex.quote(stdout[0])

    if use_sudo_password is True:
        sudo_password = host.connector_data.get("sudo_password")
        if not sudo_password:
            sudo_password = getpass("{0}sudo password: ".format(host.print_prefix))
            host.connector_data["sudo_password"] = sudo_password
        sudo_password = sudo_password
    elif callable(use_sudo_password):
        sudo_password = use_sudo_password()
    else:
        sudo_password = use_sudo_password

    return shlex.quote(sudo_password)


def remove_any_sudo_askpass_file(host):
    sudo_askpass_path = host.connector_data.get("sudo_askpass_path")
    if sudo_askpass_path:
        host.run_shell_command("rm -f {0}".format(sudo_askpass_path))
        host.connector_data["sudo_askpass_path"] = None


@memoize
def _show_use_su_login_warning():
    logger.warning(
        (
            "Using `use_su_login` may not work: "
            "some systems (MacOS, OpenBSD) ignore the flag when executing a command, "
            "use `sudo` + `use_sudo_login` instead."
        ),
    )


def make_unix_command_for_host(state, host, *command_args, **command_kwargs):
    use_sudo_password = command_kwargs.pop("use_sudo_password", None)
    if use_sudo_password:
        command_kwargs["sudo_password"] = _get_sudo_password(host, use_sudo_password)
        command_kwargs["sudo_askpass_path"] = host.connector_data.get("sudo_askpass_path")

    return make_unix_command(*command_args, **command_kwargs)


def make_unix_command(
    command,
    env=None,
    chdir=None,
    shell_executable="sh",
    # Su config
    su_user=None,
    use_su_login=False,
    su_shell=None,
    preserve_su_env=False,
    # Sudo config
    sudo=False,
    sudo_user=None,
    use_sudo_login=False,
    sudo_password=False,
    sudo_askpass_path=None,
    preserve_sudo_env=False,
    # Doas config
    doas=False,
    doas_user=None,
):
    """
    Builds a shell command with various kwargs.
    """

    if shell_executable is not None and not isinstance(shell_executable, str):
        shell_executable = "sh"

    if isinstance(command, bytes):
        command = command.decode("utf-8")

    if env:
        env_string = " ".join(['"{0}={1}"'.format(key, value) for key, value in env.items()])
        command = StringCommand("export", env_string, "&&", command)

    if chdir:
        command = StringCommand("cd", chdir, "&&", command)

    command_bits = []

    if doas:
        command_bits.extend(["doas", "-n"])

        if doas_user:
            command_bits.extend(["-u", doas_user])

    if sudo_password and sudo_askpass_path:
        command_bits.extend(
            [
                "env",
                "SUDO_ASKPASS={0}".format(sudo_askpass_path),
                MaskString("{0}={1}".format(SUDO_ASKPASS_ENV_VAR, sudo_password)),
            ],
        )

    if sudo:
        command_bits.extend(["sudo", "-H"])

        if sudo_password:
            command_bits.extend(["-A", "-k"])  # use askpass, disable cache
        else:
            command_bits.append("-n")  # disable prompt/interactivity

        if use_sudo_login:
            command_bits.append("-i")

        if preserve_sudo_env:
            command_bits.append("-E")

        if sudo_user:
            command_bits.extend(("-u", sudo_user))

    if su_user:
        command_bits.append("su")

        if use_su_login:
            _show_use_su_login_warning()
            command_bits.append("-l")

        if preserve_su_env:
            command_bits.append("-m")

        if su_shell:
            command_bits.extend(["-s", "`which {0}`".format(su_shell)])

        command_bits.extend([su_user, "-c"])

        if shell_executable is not None:
            # Quote the whole shell -c 'command' as BSD `su` does not have a shell option
            command_bits.append(
                QuoteString(StringCommand(shell_executable, "-c", QuoteString(command))),
            )
        else:
            command_bits.append(QuoteString(StringCommand(command)))
    else:
        if shell_executable is not None:
            command_bits.extend([shell_executable, "-c", QuoteString(command)])
        else:
            command_bits.extend([command])

    return StringCommand(*command_bits)


def make_win_command(command):
    """
    Builds a windows command with various kwargs.
    """

    # Quote the command as a string
    command = shlex.quote(str(command))
    command = "{0}".format(command)

    return command



"""
**Note**: this connector is in beta!

The ``@dockerssh`` connector allows you to run commands on Docker containers on a remote machine.

.. code:: shell

    # A Docker base image must be provided
    pyinfra @dockerssh/remotehost:alpine:3.8 ...

    # pyinfra can run on multiple Docker images in parallel
    pyinfra @dockerssh/remotehost:alpine:3.8,@dockerssh/remotehost:ubuntu:bionic ...
"""

import os
from tempfile import mkstemp
from typing import TYPE_CHECKING

import click

from pyinfra import logger
from pyinfra.api import QuoteString, StringCommand
from pyinfra.api.connectors import BaseConnectorMeta
from pyinfra.api.exceptions import ConnectError, InventoryError, PyinfraError
from pyinfra.api.util import get_file_io, memoize
from pyinfra.progress import progress_spinner

from . import ssh
from .util import make_unix_command_for_host

if TYPE_CHECKING:
    from pyinfra.api.host import Host
    from pyinfra.api.state import State


class Meta(BaseConnectorMeta):
    handles_execution = True


def remote_remove(
    state: "State", host: "Host", filename, print_output: bool = False, print_input: bool = False
):
    """
    Deletes a file on a remote machine over ssh.
    """
    remove_status, _, remove_stderr = ssh.run_shell_command(
        state,
        host,
        "rm -f {0}".format(filename),
        print_output=print_output,
        print_input=print_input,
    )

    if not remove_status:
        raise IOError("\n".join(remove_stderr))


@memoize
def show_warning():
    logger.warning("The @dockerssh connector is in beta!")


def make_names_data(host_image_str):
    try:
        hostname, image = host_image_str.split(":", 1)
    except (AttributeError, ValueError):  # failure to parse the host_image_str
        raise InventoryError("No ssh host or docker base image provided!")

    if not image:
        raise InventoryError("No docker base image provided!")

    show_warning()

    yield (
        "@dockerssh/{0}:{1}".format(hostname, image),
        {"ssh_hostname": hostname, "docker_image": image},
        ["@dockerssh"],
    )


def connect(state: "State", host: "Host"):
    if not host.connection:
        host.connection = ssh.connect(state, host)

    if "docker_container_id" in host.host_data:  # user can provide a docker_container_id
        return host.connection

    try:
        with progress_spinner({"docker run"}):
            # last line is the container ID
            status, stdout, stderr = ssh.run_shell_command(
                state,
                host,
                "docker run -d {0} tail -f /dev/null".format(host.data.docker_image),
            )
            if not status:
                raise IOError("\n".join(stderr))
            container_id = stdout[-1]

    except PyinfraError as e:
        host.connection = None  # fail connection
        raise ConnectError(e.args[0])

    host.host_data["docker_container_id"] = container_id
    return host.connection


def disconnect(state: "State", host: "Host"):
    container_id = host.host_data["docker_container_id"][:12]

    with progress_spinner({"docker commit"}):
        image_id = ssh.run_shell_command(state, host, "docker commit {0}".format(container_id))[1][
            -1
        ][
            7:19
        ]  # last line is the image ID, get sha256:[XXXXXXXXXX]...

    with progress_spinner({"docker rm"}):
        ssh.run_shell_command(
            state,
            host,
            "docker rm -f {0}".format(container_id),
        )

    logger.info(
        "{0}docker build complete, image ID: {1}".format(
            host.print_prefix,
            click.style(image_id, bold=True),
        ),
    )


def run_shell_command(
    state: "State",
    host: "Host",
    command,
    get_pty: bool = False,
    timeout=None,
    stdin=None,
    success_exit_codes=None,
    print_output: bool = False,
    print_input: bool = False,
    return_combined_output=False,
    **command_kwargs,
):
    container_id = host.host_data["docker_container_id"]

    # Don't sudo/su in Docker - is this the right thing to do? Makes deploys that
    # target SSH systems work w/Docker out of the box (ie most docker commands
    # are run as root).
    for key in ("sudo", "su_user"):
        command_kwargs.pop(key, None)

    command = make_unix_command_for_host(state, host, command, **command_kwargs)
    command = QuoteString(command)

    docker_flags = "-it" if get_pty else "-i"
    docker_command = StringCommand(
        "docker",
        "exec",
        docker_flags,
        container_id,
        "sh",
        "-c",
        command,
    )

    return ssh.run_shell_command(
        state,
        host,
        docker_command,
        timeout=timeout,
        stdin=stdin,
        success_exit_codes=success_exit_codes,
        print_output=print_output,
        print_input=print_input,
        return_combined_output=return_combined_output,
    )


def put_file(
    state: "State",
    host: "Host",
    filename_or_io,
    remote_filename,
    remote_temp_filename=None,
    print_output: bool = False,
    print_input: bool = False,
    **kwargs,  # ignored (sudo/etc)
):
    """
    Upload a file/IO object to the target Docker container by copying it to a
    temporary location and then uploading it into the container using ``docker cp``.
    """

    fd, local_temp_filename = mkstemp()
    remote_temp_filename = remote_temp_filename or state.get_temp_filename(local_temp_filename)

    # Load our file or IO object and write it to the temporary file
    with get_file_io(filename_or_io) as file_io:
        with open(local_temp_filename, "wb") as temp_f:
            data = file_io.read()

            if isinstance(data, str):
                data = data.encode()

            temp_f.write(data)

    # upload file to remote server
    ssh_status = ssh.put_file(state, host, local_temp_filename, remote_temp_filename)
    if not ssh_status:
        raise IOError("Failed to copy file over ssh")

    try:
        docker_id = host.host_data["docker_container_id"]
        docker_command = "docker cp {0} {1}:{2}".format(
            remote_temp_filename,
            docker_id,
            remote_filename,
        )

        status, _, stderr = ssh.run_shell_command(
            state,
            host,
            docker_command,
            print_output=print_output,
            print_input=print_input,
        )
    finally:
        os.close(fd)
        os.remove(local_temp_filename)
        remote_remove(
            state,
            host,
            local_temp_filename,
            print_output=print_output,
            print_input=print_input,
        )

    if not status:
        raise IOError("\n".join(stderr))

    if print_output:
        click.echo(
            "{0}file uploaded to container: {1}".format(
                host.print_prefix,
                remote_filename,
            ),
            err=True,
        )

    return status


def get_file(
    state: "State",
    host: "Host",
    remote_filename,
    filename_or_io,
    remote_temp_filename=None,
    print_output: bool = False,
    print_input: bool = False,
    **kwargs,  # ignored (sudo/etc)
):
    """
    Download a file from the target Docker container by copying it to a temporary
    location and then reading that into our final file/IO object.
    """

    remote_temp_filename = remote_temp_filename or state.get_temp_filename(remote_filename)

    try:
        docker_id = host.host_data["docker_container_id"]
        docker_command = "docker cp {0}:{1} {2}".format(
            docker_id,
            remote_filename,
            remote_temp_filename,
        )

        status, _, stderr = ssh.run_shell_command(
            state,
            host,
            docker_command,
            print_output=print_output,
            print_input=print_input,
        )

        ssh_status = ssh.get_file(state, host, remote_temp_filename, filename_or_io)
    finally:
        remote_remove(
            state,
            host,
            remote_temp_filename,
            print_output=print_output,
            print_input=print_input,
        )

    if not ssh_status:
        raise IOError("failed to copy file over ssh")

    if not status:
        raise IOError("\n".join(stderr))

    if print_output:
        click.echo(
            "{0}file downloaded from container: {1}".format(
                host.print_prefix,
                remote_filename,
            ),
            err=True,
        )

    return status



from unittest import TestCase

from pyinfra import host, inventory
from pyinfra.api import Host, Inventory
from pyinfra.context import ctx_host, ctx_inventory


def _create_host():
    return Host(None, None, None, None)


class TestHostContextObject(TestCase):
    def test_context_host_dir_matches_class(self):
        assert dir(Host) == dir(host)

    def test_context_host_eq(self):
        host_obj = _create_host()
        ctx_host.set(host_obj)
        assert host == host_obj

    def test_context_host_repr(self):
        host_obj = _create_host()
        ctx_host.set(host_obj)
        assert repr(host) == "ContextObject(Host):Host(None)"

    def test_context_host_str(self):
        host_obj = _create_host()
        ctx_host.set(host_obj)
        assert str(host_obj) == "None"

    def test_context_host_attr(self):
        host_obj = _create_host()
        ctx_host.set(host_obj)

        with self.assertRaises(AttributeError):
            host_obj.hello

        setattr(host, "hello", "world")
        assert host_obj.hello == host.hello

    def test_context_host_class_attr(self):
        host_obj = _create_host()
        ctx_host.set(host_obj)
        assert ctx_host.isset() is True

        with self.assertRaises(AttributeError):
            host_obj.hello

        setattr(Host, "hello", "class_world")
        setattr(host_obj, "hello", "instance_world")

        assert host.hello == host.hello

        # Reset and check fallback to class variable
        ctx_host.reset()
        assert ctx_host.isset() is False
        assert host.hello == "class_world"


class TestInventoryContextObject(TestCase):
    def test_context_inventory_len(self):
        inventory_obj = Inventory(("host", "anotherhost"))
        ctx_inventory.set(inventory_obj)
        assert ctx_inventory.isset() is True

        assert len(inventory) == len(inventory_obj)

    def test_context_inventory_iter(self):
        inventory_obj = Inventory(("host", "anotherhost"))
        ctx_inventory.set(inventory_obj)
        assert ctx_inventory.isset() is True

        assert list(iter(inventory)) == list(iter(inventory_obj))



import json
from unittest import TestCase
from unittest.mock import mock_open, patch

from pyinfra.api.exceptions import InventoryError
from pyinfra.connectors.mech import get_mech_options, make_names_data

FAKE_MECH_OPTIONS = {
    "groups": {
        "ubuntu16": ["mygroup"],
    },
    "data": {
        "centos7": {
            "somedata": "somevalue",
        },
    },
}
FAKE_MECH_OPTIONS_DATA = json.dumps(FAKE_MECH_OPTIONS)


def fake_mech_shell(command, splitlines=None):
    if command == "mech ls":
        return [
            "NAME	        ADDRESS	                                BOX	     VERSION	PATH",  # noqa: E501
            "ubuntu16	       192.168.2.226             bento/ubuntu-16.04	 201912.04.0	/Users/bob/somedir/ubuntu16",  # noqa: E501
            "centos7	       192.168.2.227                 bento/centos-7	 201912.05.0	/Users/bob/somedir/centos7",  # noqa: E501
            "centos6	       poweroff                      bento/centos-6	 201912.04.0	/Users/bob/somedir/centos6",  # noqa: E501
            "fedora31	               	                    bento/fedora-31	 201912.04.0	/Users/bob/somedir/fedora31",  # noqa: E501
        ]
    elif command == "mech ssh-config ubuntu16":
        return [
            "Host ubuntu16",
            "  User vagrant",
            "  Port 22",
            "  UserKnownHostsFile /dev/null",
            "  StrictHostKeyChecking no",
            "  PasswordAuthentication no",
            "  IdentityFile path/to/key",
            "  IdentitiesOnly yes",
            "  LogLevel FATAL",
            "  HostName 192.168.2.226",
            "",
        ]
    elif command == "mech ssh-config centos7":
        return [
            "Host centos7",
            "  User vagrant",
            "  Port 22",
            "  UserKnownHostsFile /dev/null",
            "  StrictHostKeyChecking no",
            "  PasswordAuthentication no",
            "  IdentityFile path/to/key",
            "  IdentitiesOnly yes",
            "  LogLevel FATAL",
            "  HostName 192.168.2.227",
            "",
        ]
    elif command == "mech ssh-config centos6":
        return [
            "ERROR: Error: The virtual machine is not powered on: /Users/bob/debian8/.mech/debian-8.11-amd64.vmx",  # noqa: E501
            "This Mech machine is reporting that it is not yet ready for SSH. Make",
            "sure your machine is created and running and try again. Additionally,",
            "check the output of `mech status` to verify that the machine is in the",
            "state that you expect.",
        ]

    return []


@patch("pyinfra.connectors.mech.local.shell", fake_mech_shell)
class TestMechConnector(TestCase):
    def tearDown(self):
        get_mech_options.cache = {}

    @patch(
        "pyinfra.connectors.mech.open",
        mock_open(read_data=FAKE_MECH_OPTIONS_DATA),
        create=True,
    )
    @patch("pyinfra.connectors.mech.path.exists", lambda path: True)
    def test_make_names_data_with_options(self):
        data = make_names_data()

        assert data == [
            (
                "@mech/ubuntu16",
                {
                    "ssh_port": "22",
                    "ssh_user": "vagrant",
                    "ssh_hostname": "192.168.2.226",
                    "ssh_key": "path/to/key",
                },
                ["mygroup", "@mech"],
            ),
            (
                "@mech/centos7",
                {
                    "ssh_port": "22",
                    "ssh_user": "vagrant",
                    "ssh_hostname": "192.168.2.227",
                    "ssh_key": "path/to/key",
                    "somedata": "somevalue",
                },
                ["@mech"],
            ),
        ]

    def test_make_names_data_with_limit(self):
        data = make_names_data(limit=("ubuntu16",))

        assert data == [
            (
                "@mech/ubuntu16",
                {
                    "ssh_port": "22",
                    "ssh_user": "vagrant",
                    "ssh_hostname": "192.168.2.226",
                    "ssh_key": "path/to/key",
                },
                ["@mech"],
            ),
        ]

    def test_make_names_data_no_matches(self):
        with self.assertRaises(InventoryError):
            make_names_data(limit="nope")



"""
Connect to hosts over SSH. This is the default connector and all targets default
to this meaning you do not need to specify it - ie the following two commands
are identical:

.. code:: shell

    pyinfra my-host.net ...
    pyinfra @ssh/my-host.net ...
"""

from distutils.spawn import find_executable
from getpass import getpass
from os import path
from socket import error as socket_error, gaierror
from typing import TYPE_CHECKING, Type, Union

import click
from paramiko import (
    AuthenticationException,
    BadHostKeyException,
    DSSKey,
    ECDSAKey,
    Ed25519Key,
    PasswordRequiredException,
    RSAKey,
    SFTPClient,
    SSHException,
)

import pyinfra
from pyinfra import logger
from pyinfra.api.command import QuoteString, StringCommand
from pyinfra.api.connectors import BaseConnectorMeta
from pyinfra.api.exceptions import ConnectError, PyinfraError
from pyinfra.api.util import get_file_io, memoize

from .sshuserclient import SSHClient
from .util import (
    execute_command_with_sudo_retry,
    make_unix_command_for_host,
    read_buffers_into_queue,
    run_local_process,
    split_combined_output,
    write_stdin,
)

if TYPE_CHECKING:
    from pyinfra.api.host import Host
    from pyinfra.api.state import State


class Meta(BaseConnectorMeta):
    handles_execution = True
    keys_prefix = "ssh"

    class DataKeys:
        hostname = "SSH hostname"
        port = "SSH port"

        user = "User to SSH as"
        password = "Password to use for authentication"
        key = "Key file to use for authentication"
        key_password = "Key file password"

        allow_agent = "Allow using SSH agent"
        look_for_keys = "Allow looking up users keys"

        forward_agent = "Enable SSH forward agent"
        config_file = "Custom SSH config file"
        known_hosts_file = "Custom SSH known hosts file"
        strict_host_key_checking = "Override strict host keys check setting"

        paramiko_connect_kwargs = (
            "Override keyword arguments passed into paramiko's `SSHClient.connect`"
        )


DATA_KEYS = Meta.keys()


def make_names_data(hostname):
    yield "@ssh/{0}".format(hostname), {DATA_KEYS.hostname: hostname}, []


def _raise_connect_error(host: "Host", message, data):
    message = "{0} ({1})".format(message, data)
    raise ConnectError(message)


def _load_private_key_file(filename: str, key_filename: str, key_password: str):
    exception: Union[PyinfraError, SSHException] = PyinfraError("Invalid key: {0}".format(filename))

    key_cls: Union[Type[RSAKey], Type[DSSKey], Type[ECDSAKey], Type[Ed25519Key]]

    for key_cls in (RSAKey, DSSKey, ECDSAKey, Ed25519Key):
        try:
            return key_cls.from_private_key_file(
                filename=filename,
            )

        except PasswordRequiredException:
            if not key_password:
                # If password is not provided, but we're in CLI mode, ask for it. I'm not a
                # huge fan of having CLI specific code in here, but it doesn't really fit
                # anywhere else without duplicating lots of key related code into cli.py.
                if pyinfra.is_cli:
                    key_password = getpass(
                        "Enter password for private key: {0}: ".format(
                            key_filename,
                        ),
                    )

                # API mode and no password? We can't continue!
                else:
                    raise PyinfraError(
                        "Private key file ({0}) is encrypted, set ssh_key_password to "
                        "use this key".format(key_filename),
                    )

            try:
                return key_cls.from_private_key_file(
                    filename=filename,
                    password=key_password,
                )
            except SSHException as e:  # key does not match key_cls type
                exception = e
        except SSHException as e:  # key does not match key_cls type
            exception = e
    raise exception


def _get_private_key(state: "State", key_filename: str, key_password: str):
    if key_filename in state.private_keys:
        return state.private_keys[key_filename]

    ssh_key_filenames = [
        # Global from executed directory
        path.expanduser(key_filename),
    ]

    if state.cwd:
        # Relative to the CWD
        path.join(state.cwd, key_filename)

    key = None
    key_file_exists = False

    for filename in ssh_key_filenames:
        if not path.isfile(filename):
            continue

        key_file_exists = True

        try:
            key = _load_private_key_file(filename, key_filename, key_password)
            break
        except SSHException:
            pass

    # No break, so no key found
    if not key:
        if not key_file_exists:
            raise PyinfraError("No such private key file: {0}".format(key_filename))
        raise PyinfraError("Invalid private key file: {0}".format(key_filename))

    # Load any certificate, names from OpenSSH:
    # https://github.com/openssh/openssh-portable/blob/049297de975b92adcc2db77e3fb7046c0e3c695d/ssh-keygen.c#L2453  # noqa: E501
    for certificate_filename in (
        "{0}-cert.pub".format(key_filename),
        "{0}.pub".format(key_filename),
    ):
        if path.isfile(certificate_filename):
            key.load_certificate(certificate_filename)

    state.private_keys[key_filename] = key
    return key


def _make_paramiko_kwargs(state: "State", host: "Host"):
    kwargs = {
        "allow_agent": False,
        "look_for_keys": False,
        "hostname": host.data.get(DATA_KEYS.hostname, host.name),
        # Overrides of SSH config via pyinfra host data
        "_pyinfra_ssh_forward_agent": host.data.get(DATA_KEYS.forward_agent),
        "_pyinfra_ssh_config_file": host.data.get(DATA_KEYS.config_file),
        "_pyinfra_ssh_known_hosts_file": host.data.get(DATA_KEYS.known_hosts_file),
        "_pyinfra_ssh_strict_host_key_checking": host.data.get(DATA_KEYS.strict_host_key_checking),
        "_pyinfra_ssh_paramiko_connect_kwargs": host.data.get(DATA_KEYS.paramiko_connect_kwargs),
    }

    for key, value in (
        ("username", host.data.get(DATA_KEYS.user)),
        ("port", int(host.data.get(DATA_KEYS.port, 0))),
        ("timeout", state.config.CONNECT_TIMEOUT),
    ):
        if value:
            kwargs[key] = value

    # Password auth (boo!)
    ssh_password = host.data.get(DATA_KEYS.password)
    if ssh_password:
        kwargs["password"] = ssh_password

    # Key auth!
    ssh_key = host.data.get(DATA_KEYS.key)
    if ssh_key:
        kwargs["pkey"] = _get_private_key(
            state,
            key_filename=ssh_key,
            key_password=host.data.get(DATA_KEYS.key_password),
        )

    # No key or password, so let's have paramiko look for SSH agents and user keys
    # unless disabled by the user.
    else:
        kwargs["allow_agent"] = host.data.get(DATA_KEYS.allow_agent, True)
        kwargs["look_for_keys"] = host.data.get(DATA_KEYS.look_for_keys, True)

    return kwargs


def connect(state: "State", host: "Host"):
    """
    Connect to a single host. Returns the SSH client if successful. Stateless by
    design so can be run in parallel.
    """

    kwargs = _make_paramiko_kwargs(state, host)
    logger.debug("Connecting to: %s (%r)", host.name, kwargs)

    hostname = kwargs.pop("hostname")

    try:
        # Create new client & connect to the host
        client = SSHClient()
        client.connect(hostname, **kwargs)
        return client

    except AuthenticationException as e:
        auth_kwargs = {}

        for key, value in kwargs.items():
            if key in ("username", "password"):
                auth_kwargs[key] = value
                continue

            if key == "pkey" and value:
                auth_kwargs["key"] = host.data.get(DATA_KEYS.key)

        auth_args = ", ".join("{0}={1}".format(key, value) for key, value in auth_kwargs.items())

        _raise_connect_error(host, "Authentication error ({0})".format(auth_args), e)

    except BadHostKeyException as e:
        remove_entry = e.hostname
        port = client._ssh_config.get("port", 22)
        if port != 22:
            remove_entry = f"[{e.hostname}]:{port}"

        logger.warning("WARNING: REMOTE HOST IDENTIFICATION HAS CHANGED!")
        logger.warning(
            ("Someone could be eavesdropping on you right now " "(man-in-the-middle attack)!"),
        )
        logger.warning("If this is expected, you can remove the bad key using:")
        logger.warning(f"    ssh-keygen -R {remove_entry}")

        _raise_connect_error(
            host,
            "SSH host key error",
            f"Host key for {e.hostname} does not match.",
        )

    except SSHException as e:
        _raise_connect_error(host, "SSH error", e)

    except gaierror:
        _raise_connect_error(host, "Could not resolve hostname", hostname)

    except socket_error as e:
        _raise_connect_error(host, "Could not connect", e)

    except EOFError as e:
        _raise_connect_error(host, "EOF error", e)


def run_shell_command(
    state: "State",
    host: "Host",
    command,
    get_pty: bool = False,
    timeout=None,
    stdin=None,
    success_exit_codes=None,
    print_output: bool = False,
    print_input: bool = False,
    return_combined_output: bool = False,
    **command_kwargs,
):
    """
    Execute a command on the specified host.

    Args:
        state (``pyinfra.api.State`` obj): state object for this command
        hostname (string): hostname of the target
        command (string): actual command to execute
        sudo (boolean): whether to wrap the command with sudo
        sudo_user (string): user to sudo to
        get_pty (boolean): whether to get a PTY before executing the command
        env (dict): environment variables to set
        timeout (int): timeout for this command to complete before erroring

    Returns:
        tuple: (exit_code, stdout, stderr)
        stdout and stderr are both lists of strings from each buffer.
    """

    def execute_command():
        unix_command = make_unix_command_for_host(state, host, command, **command_kwargs)
        actual_command = unix_command.get_raw_value()

        logger.debug(
            "Running command on %s: (pty=%s) %s",
            host.name,
            get_pty,
            unix_command,
        )

        if print_input:
            click.echo("{0}>>> {1}".format(host.print_prefix, unix_command), err=True)

        # Run it! Get stdout, stderr & the underlying channel
        stdin_buffer, stdout_buffer, stderr_buffer = host.connection.exec_command(
            actual_command,
            get_pty=get_pty,
        )

        if stdin:
            write_stdin(stdin, stdin_buffer)

        combined_output = read_buffers_into_queue(
            stdout_buffer,
            stderr_buffer,
            timeout=timeout,
            print_output=print_output,
            print_prefix=host.print_prefix,
        )

        logger.debug("Waiting for exit status...")
        exit_status = stdout_buffer.channel.recv_exit_status()
        logger.debug("Command exit status: %i", exit_status)

        return exit_status, combined_output

    return_code, combined_output = execute_command_with_sudo_retry(
        host,
        command_kwargs,
        execute_command,
    )

    if success_exit_codes:
        status = return_code in success_exit_codes
    else:
        status = return_code == 0

    if return_combined_output:
        return status, combined_output

    stdout, stderr = split_combined_output(combined_output)
    return status, stdout, stderr


@memoize
def _get_sftp_connection(host: "Host"):
    assert host.connection is not None
    transport = host.connection.get_transport()

    try:
        return SFTPClient.from_transport(transport)
    except SSHException as e:
        raise ConnectError(
            (
                "Unable to establish SFTP connection. Check that the SFTP subsystem "
                "for the SSH service at {0} is enabled."
            ).format(host),
        ) from e


def _get_file(host: "Host", remote_filename: str, filename_or_io):
    with get_file_io(filename_or_io, "wb") as file_io:
        sftp = _get_sftp_connection(host)
        sftp.getfo(remote_filename, file_io)


def get_file(
    state: "State",
    host: "Host",
    remote_filename: str,
    filename_or_io,
    remote_temp_filename=None,
    sudo: bool = False,
    sudo_user=None,
    su_user=None,
    print_output: bool = False,
    print_input: bool = False,
    **command_kwargs,
):
    """
    Download a file from the remote host using SFTP. Supports download files
    with sudo by copying to a temporary directory with read permissions,
    downloading and then removing the copy.
    """

    if sudo or su_user:
        # Get temp file location
        temp_file = remote_temp_filename or state.get_temp_filename(remote_filename)

        # Copy the file to the tempfile location and add read permissions
        command = "cp {0} {1} && chmod +r {0}".format(remote_filename, temp_file)

        copy_status, _, stderr = run_shell_command(
            state,
            host,
            command,
            sudo=sudo,
            sudo_user=sudo_user,
            su_user=su_user,
            print_output=print_output,
            print_input=print_input,
            **command_kwargs,
        )

        if copy_status is False:
            logger.error("File download copy temp error: {0}".format("\n".join(stderr)))
            return False

        try:
            _get_file(host, temp_file, filename_or_io)

        # Ensure that, even if we encounter an error, we (attempt to) remove the
        # temporary copy of the file.
        finally:
            remove_status, _, stderr = run_shell_command(
                state,
                host,
                "rm -f {0}".format(temp_file),
                sudo=sudo,
                sudo_user=sudo_user,
                su_user=su_user,
                print_output=print_output,
                print_input=print_input,
                **command_kwargs,
            )

        if remove_status is False:
            logger.error("File download remove temp error: {0}".format("\n".join(stderr)))
            return False

    else:
        _get_file(host, remote_filename, filename_or_io)

    if print_output:
        click.echo(
            "{0}file downloaded: {1}".format(host.print_prefix, remote_filename),
            err=True,
        )

    return True


def _put_file(host: "Host", filename_or_io, remote_location):
    attempts = 1

    try:
        with get_file_io(filename_or_io) as file_io:
            sftp = _get_sftp_connection(host)
            sftp.putfo(file_io, remote_location)
    except OSError as e:
        if attempts > 3:
            raise
        logger.warning(f"Failed to upload file, retrying: {e}")
        attempts += 1


def put_file(
    state: "State",
    host: "Host",
    filename_or_io,
    remote_filename,
    remote_temp_filename=None,
    sudo: bool = False,
    sudo_user=None,
    su_user=None,
    print_output: bool = False,
    print_input: bool = False,
    **command_kwargs,
):
    """
    Upload file-ios to the specified host using SFTP. Supports uploading files
    with sudo by uploading to a temporary directory then moving & chowning.
    """

    # sudo/su are a little more complicated, as you can only sftp with the SSH
    # user connected, so upload to tmp and copy/chown w/sudo and/or su_user
    if sudo or su_user:
        # Get temp file location
        temp_file = remote_temp_filename or state.get_temp_filename(remote_filename)
        _put_file(host, filename_or_io, temp_file)

        # Make sure our sudo/su user can access the file
        if su_user:
            command = StringCommand("setfacl", "-m", "u:{0}:r".format(su_user), temp_file)
        elif sudo_user:
            command = StringCommand("setfacl -m u:{0}:r".format(sudo_user), temp_file)
        if su_user or sudo_user:
            status, _, stderr = run_shell_command(
                state,
                host,
                command,
                sudo=False,
                print_output=print_output,
                print_input=print_input,
                **command_kwargs,
            )

            if status is False:
                logger.error("Error on handover to sudo/su user: {0}".format("\n".join(stderr)))
                return False

        # Execute run_shell_command w/sudo and/or su_user
        command = StringCommand("cp", temp_file, QuoteString(remote_filename))

        status, _, stderr = run_shell_command(
            state,
            host,
            command,
            sudo=sudo,
            sudo_user=sudo_user,
            su_user=su_user,
            print_output=print_output,
            print_input=print_input,
            **command_kwargs,
        )

        if status is False:
            logger.error("File upload error: {0}".format("\n".join(stderr)))
            return False

        # Delete the temporary file now that we've successfully copied it
        command = StringCommand("rm", "-f", temp_file)

        status, _, stderr = run_shell_command(
            state,
            host,
            command,
            sudo=False,
            print_output=print_output,
            print_input=print_input,
            **command_kwargs,
        )

        if status is False:
            logger.error("Unable to remove temporary file: {0}".format("\n".join(stderr)))
            return False

    # No sudo and no su_user, so just upload it!
    else:
        _put_file(host, filename_or_io, remote_filename)

    if print_output:
        click.echo(
            "{0}file uploaded: {1}".format(host.print_prefix, remote_filename),
            err=True,
        )

    return True


def check_can_rsync(host: "Host"):
    if host.data.get(DATA_KEYS.key_password):
        raise NotImplementedError("Rsync does not currently work with SSH keys needing passwords.")

    if host.data.get(DATA_KEYS.password):
        raise NotImplementedError("Rsync does not currently work with SSH passwords.")

    if not find_executable("rsync"):
        raise NotImplementedError("The `rsync` binary is not available on this system.")


def rsync(
    state: "State",
    host: "Host",
    src: str,
    dest: str,
    flags,
    print_output: bool = False,
    print_input: bool = False,
    sudo: bool = False,
    sudo_user=None,
    **ignored_kwargs,
):
    hostname = host.data.get(DATA_KEYS.hostname, host.name)
    user = host.data.get(DATA_KEYS.user, "")
    if user:
        user = "{0}@".format(user)

    ssh_flags = []

    port = host.data.get(DATA_KEYS.port)
    if port:
        ssh_flags.append("-p {0}".format(port))

    ssh_key = host.data.get(DATA_KEYS.key)
    if ssh_key:
        ssh_flags.append("-i {0}".format(ssh_key))

    remote_rsync_command = "rsync"
    if sudo:
        remote_rsync_command = "sudo rsync"
        if sudo_user:
            remote_rsync_command = "sudo -u {0} rsync".format(sudo_user)

    # To avoid asking for interactive input, specify BatchMode=yes
    rsync_command = (
        "rsync {rsync_flags} "
        "--rsh 'ssh -o BatchMode=yes {ssh_flags}' "
        "--rsync-path '{remote_rsync_command}' "
        "{src} {user}{hostname}:{dest}"
    ).format(
        rsync_flags=" ".join(flags),
        ssh_flags=" ".join(ssh_flags),
        remote_rsync_command=remote_rsync_command,
        user=user,
        hostname=hostname,
        src=src,
        dest=dest,
    )

    if print_input:
        click.echo("{0}>>> {1}".format(host.print_prefix, rsync_command), err=True)

    return_code, combined_output = run_local_process(
        rsync_command,
        print_output=print_output,
        print_prefix=host.print_prefix,
    )

    status = return_code == 0

    if not status:
        _, stderr = split_combined_output(combined_output)
        raise IOError("\n".join(stderr))

    return True



import re

from pyinfra.api.facts import FactBase


class GitBranch(FactBase):
    requires_command = "git"

    @staticmethod
    def command(repo):
        return "! test -d {0} || (cd {0} && git describe --all)".format(repo)

    @staticmethod
    def process(output):
        return re.sub(r"(heads|tags)/", r"", "\n".join(output))


class GitConfig(FactBase):
    default = dict

    requires_command = "git"

    @staticmethod
    def command(repo=None):
        if repo is None:
            return "git config --global -l || true"

        return "! test -d {0} || (cd {0} && git config --local -l)".format(repo)

    @staticmethod
    def process(output):
        items = {}

        for line in output:
            key, value = line.split("=", 1)
            items[key] = value

        return items


class GitTrackingBranch(FactBase):
    requires_command = "git"

    @staticmethod
    def command(repo):
        return r"! test -d {0} || (cd {0} && git status --branch --porcelain)".format(repo)

    @staticmethod
    def process(output):
        if not output:
            return None

        m = re.search(r"\.{3}(\S+)\b", output[0])
        if m:
            return m.group(1)
        return None



"""
Manage apt packages and repositories.
"""

from datetime import datetime, timedelta
from urllib.parse import urlparse

from pyinfra import host, state
from pyinfra.api import OperationError, operation
from pyinfra.facts.apt import AptKeys, AptSources, parse_apt_repo
from pyinfra.facts.deb import DebPackage, DebPackages
from pyinfra.facts.files import File
from pyinfra.facts.gpg import GpgKey
from pyinfra.facts.server import Date

from . import files
from .util.packaging import ensure_packages

APT_UPDATE_FILENAME = "/var/lib/apt/periodic/update-success-stamp"


def noninteractive_apt(command, force=False):
    args = ["DEBIAN_FRONTEND=noninteractive apt-get -y"]

    if force:
        args.append("--force-yes")

    args.extend(
        (
            '-o Dpkg::Options::="--force-confdef"',
            '-o Dpkg::Options::="--force-confold"',
            command,
        ),
    )

    return " ".join(args)


@operation
def key(src=None, keyserver=None, keyid=None):
    """
    Add apt gpg keys with ``apt-key``.

    + src: filename or URL
    + keyserver: URL of keyserver to fetch key from
    + keyid: key ID or list of key IDs when using keyserver

    keyserver/id:
        These must be provided together.

    **Examples:**

    .. code:: python

        # Note: If using URL, wget is assumed to be installed.
        apt.key(
            name="Add the Docker apt gpg key",
            src="https://download.docker.com/linux/ubuntu/gpg",
        )

        apt.key(
            name="Install VirtualBox key",
            src="https://www.virtualbox.org/download/oracle_vbox_2016.asc",
        )
    """

    existing_keys = host.get_fact(AptKeys)

    if src:
        key_data = host.get_fact(GpgKey, src=src)
        if key_data:
            keyid = list(key_data.keys())

        if not keyid or not all(kid in existing_keys for kid in keyid):
            # If URL, wget the key to stdout and pipe into apt-key, because the "adv"
            # apt-key passes to gpg which doesn't always support https!
            if urlparse(src).scheme:
                yield "(wget -O - {0} || curl -sSLf {0}) | apt-key add -".format(src)
            else:
                yield "apt-key add {0}".format(src)

            if keyid:
                for kid in keyid:
                    existing_keys[kid] = {}
        else:
            host.noop("All keys from {0} are already available in the apt keychain".format(src))

    if keyserver:
        if not keyid:
            raise OperationError("`keyid` must be provided with `keyserver`")

        if isinstance(keyid, str):
            keyid = [keyid]

        needed_keys = sorted(set(keyid) - set(existing_keys.keys()))
        if needed_keys:
            yield "apt-key adv --keyserver {0} --recv-keys {1}".format(
                keyserver,
                " ".join(needed_keys),
            )
            for kid in keyid:
                existing_keys[kid] = {}
        else:
            host.noop(
                "Keys {0} are already available in the apt keychain".format(
                    ", ".join(keyid),
                ),
            )


@operation
def repo(src, present=True, filename=None):
    """
    Add/remove apt repositories.

    + src: apt source string eg ``deb http://X hardy main``
    + present: whether the repo should exist on the system
    + filename: optional filename to use ``/etc/apt/sources.list.d/<filename>.list``. By
      default uses ``/etc/apt/sources.list``.

    **Example:**

    .. code:: python

        apt.repo(
            name="Install VirtualBox repo",
            src="deb https://download.virtualbox.org/virtualbox/debian bionic contrib",
        )
    """

    # Get the target .list file to manage
    if filename:
        filename = "/etc/apt/sources.list.d/{0}.list".format(filename)
    else:
        filename = "/etc/apt/sources.list"

    # Work out if the repo exists already
    apt_sources = host.get_fact(AptSources)

    is_present = False
    repo = parse_apt_repo(src)
    if repo and repo in apt_sources:
        is_present = True

    # Doesn't exist and we want it
    if not is_present and present:
        yield from files.line(
            filename,
            src,
            escape_regex_characters=True,
        )
        apt_sources.append(repo)

    # Exists and we don't want it
    elif is_present and not present:
        yield from files.line(
            filename,
            src,
            present=False,
            assume_present=True,
            escape_regex_characters=True,
        )
        apt_sources.remove(repo)

    else:
        host.noop(
            'apt repo "{0}" {1}'.format(
                src,
                "exists" if present else "does not exist",
            ),
        )


@operation(is_idempotent=False)
def ppa(src, present=True):
    """
    Add/remove Ubuntu ppa repositories.

    + src: the PPA name (full ppa:user/repo format)
    + present: whether it should exist

    Note:
        requires ``apt-add-repository`` on the remote host

    **Example:**

    .. code:: python

        # Note: Assumes software-properties-common is installed.
        apt.ppa(
            name="Add the Bitcoin ppa",
            src="ppa:bitcoin/bitcoin",
        )

    """

    if present:
        yield 'apt-add-repository -y "{0}"'.format(src)

    if not present:
        yield 'apt-add-repository -y --remove "{0}"'.format(src)


@operation
def deb(src, present=True, force=False):
    """
    Add/remove ``.deb`` file packages.

    + src: filename or URL of the ``.deb`` file
    + present: whether or not the package should exist on the system
    + force: whether to force the package install by passing `--force-yes` to apt

    Note:
        When installing, ``apt-get install -f`` will be run to install any unmet
        dependencies.

    URL sources with ``present=False``:
        If the ``.deb`` file isn't downloaded, pyinfra can't remove any existing
        package as the file won't exist until mid-deploy.

    **Example:**

    .. code:: python

        # Note: Assumes wget is installed.
        apt.deb(
            name="Install Chrome via deb",
            src="https://dl.google.com/linux/direct/google-chrome-stable_current_amd64.deb",
        )
    """

    original_src = src

    # If source is a url
    if urlparse(src).scheme:
        # Generate a temp filename
        temp_filename = state.get_temp_filename(src)

        # Ensure it's downloaded
        yield from files.download(src, temp_filename)

        # Override the source with the downloaded file
        src = temp_filename

    # Check for file .deb information (if file is present)
    info = host.get_fact(DebPackage, name=src)
    current_packages = host.get_fact(DebPackages)

    exists = False

    # We have deb info! Check against installed packages
    if info:
        if (
            info["name"] in current_packages
            and info.get("version") in current_packages[info["name"]]
        ):
            exists = True

    # Package does not exist and we want?
    if present:
        if not exists:
            # Install .deb file - ignoring failure (on unmet dependencies)
            yield "dpkg --force-confdef --force-confold -i {0} 2> /dev/null || true".format(src)
            # Attempt to install any missing dependencies
            yield "{0} -f".format(noninteractive_apt("install", force=force))
            # Now reinstall, and critically configure, the package - if there are still
            # missing deps, now we error
            yield "dpkg --force-confdef --force-confold -i {0}".format(src)

            if info:
                current_packages[info["name"]] = [info.get("version")]
        else:
            host.noop("deb {0} is installed".format(original_src))

    # Package exists but we don't want?
    if not present:
        if exists:
            yield "{0} {1}".format(
                noninteractive_apt("remove", force=force),
                info["name"],
            )
            current_packages.pop(info["name"])
        else:
            host.noop("deb {0} is not installed".format(original_src))


@operation(
    is_idempotent=False,
    idempotent_notice=(
        "This operation will always execute commands "
        "unless the ``cache_time`` argument is provided."
    ),
)
def update(cache_time=None):
    """
    Updates apt repositories.

    + cache_time: cache updates for this many seconds

    **Example:**

    .. code:: python

        apt.update(
            name="Update apt repositories",
            cache_time=3600,
        )
    """

    # If cache_time check when apt was last updated, prevent updates if within time
    if cache_time:
        # Ubuntu provides this handy file
        cache_info = host.get_fact(File, path=APT_UPDATE_FILENAME)

        # Time on files is not tz-aware, and will be the same tz as the server's time,
        # so we can safely remove the tzinfo from the Date fact before comparison.
        host_cache_time = host.get_fact(Date).replace(tzinfo=None) - timedelta(seconds=cache_time)
        if cache_info and cache_info["mtime"] and cache_info["mtime"] > host_cache_time:
            host.noop("apt is already up to date")
            return

    yield "apt-get update"

    # Some apt systems (Debian) have the /var/lib/apt/periodic directory, but
    # don't bother touching anything in there - so pyinfra does it, enabling
    # cache_time to work.
    if cache_time:
        yield "touch {0}".format(APT_UPDATE_FILENAME)
        if cache_info is None:
            host.create_fact(
                File,
                kwargs={"path": APT_UPDATE_FILENAME},
                data={"mtime": datetime.utcnow()},
            )
        else:
            cache_info["mtime"] = datetime.utcnow()


_update = update  # noqa: E305


@operation(is_idempotent=False)
def upgrade():
    """
    Upgrades all apt packages.

    **Example:**

    .. code:: python

        apt.upgrade(
            name="Upgrade apt packages",
        )
    """

    yield noninteractive_apt("upgrade")


_upgrade = upgrade  # noqa: E305 (for use below where update is a kwarg)


@operation(is_idempotent=False)
def dist_upgrade():
    """
    Updates all apt packages, employing dist-upgrade.

    **Example:**

    .. code:: python

        apt.dist_upgrade(
            name="Upgrade apt packages using dist-upgrade",
        )
    """

    yield noninteractive_apt("dist-upgrade")


@operation
def packages(
    packages=None,
    present=True,
    latest=False,
    update=False,
    cache_time=None,
    upgrade=False,
    force=False,
    no_recommends=False,
    allow_downgrades=False,
    extra_install_args=None,
    extra_uninstall_args=None,
):
    """
    Install/remove/update packages & update apt.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``apt update`` before installing packages
    + cache_time: when used with ``update``, cache for this many seconds
    + upgrade: run ``apt upgrade`` before installing packages
    + force: whether to force package installs by passing `--force-yes` to apt
    + no_recommends: don't install recommended packages
    + allow_downgrades: allow downgrading packages with version (--allow-downgrades)
    + extra_install_args: additional arguments to the apt install command
    + extra_uninstall_args: additional arguments to the apt uninstall command

    Versions:
        Package versions can be pinned like apt: ``<pkg>=<version>``

    Cache time:
        When ``cache_time`` is set the ``/var/lib/apt/periodic/update-success-stamp`` file
        is touched upon successful update. Some distros already do this (Ubuntu), but others
        simply leave the periodic directory empty (Debian).

    **Examples:**

    .. code:: python

        # Update package list and install packages
        apt.packages(
            name="Install Asterisk and Vim",
            packages=["asterisk", "vim"],
            update=True,
        )

        # Install the latest versions of packages (always check)
        apt.packages(
            name="Install latest Vim",
            packages=["vim"],
            latest=True,
        )

        # Note: host.get_fact(OsVersion) is the same as `uname -r` (ex: '4.15.0-72-generic')
        apt.packages(
            name="Install kernel headers",
            packages=[f"linux-headers-{host.get_fact(OsVersion)}"],
            update=True,
        )
    """

    if update:
        yield from _update(cache_time=cache_time)

    if upgrade:
        yield from _upgrade()

    install_command = ["install"]
    if no_recommends is True:
        install_command.append("--no-install-recommends")
    if allow_downgrades:
        install_command.append("--allow-downgrades")

    upgrade_command = " ".join(install_command)

    if extra_install_args:
        install_command.append(extra_install_args)

    install_command = " ".join(install_command)

    uninstall_command = ["remove"]
    if extra_uninstall_args:
        uninstall_command.append(extra_uninstall_args)

    uninstall_command = " ".join(uninstall_command)

    # Compare/ensure packages are present/not
    yield from ensure_packages(
        host,
        packages,
        host.get_fact(DebPackages),
        present,
        install_command=noninteractive_apt(install_command, force=force),
        uninstall_command=noninteractive_apt(uninstall_command, force=force),
        upgrade_command=noninteractive_apt(upgrade_command, force=force),
        version_join="=",
        latest=latest,
    )



"""
Welcome to pyinfra.
"""

import logging

# Global flag set True by `pyinfra_cli.__main__`
is_cli = False

# Global pyinfra logger
logger = logging.getLogger("pyinfra")

# Trigger context module creation
from .context import config, host, init_base_classes, inventory, state  # noqa

# Setup package level version
from .version import __version__  # noqa

# Initialise base classes - this sets the context modules to point at the underlying
# class objects (Host, etc), which makes ipython/etc work as expected.
init_base_classes()



"""
Manage launchd services.
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.launchd import LaunchdStatus

from .util.service import handle_service_control


@operation
def service(
    service,
    running=True,
    restarted=False,
    command=None,
):
    """
    Manage the state of systemd managed services.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + command: custom command to pass like: ``launchctl <command> <service>``
    + enabled: whether this service should be enabled/disabled on boot
    """

    was_running = host.get_fact(LaunchdStatus).get(service, None)

    yield from handle_service_control(
        host,
        service,
        host.get_fact(LaunchdStatus),
        "launchctl {1} {0}",
        # No support for restart/reload/command
        running,
        None,
        None,
        None,
    )

    # No restart command, so just stop/start
    if restarted and was_running:
        yield "launchctl stop {0}".format(service)
        yield "launchctl start {0}".format(service)



"""
Manage upstart services.
"""

from io import StringIO

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.upstart import UpstartStatus

from . import files
from .util.service import handle_service_control


@operation
def service(
    service,
    running=True,
    restarted=False,
    reloaded=False,
    command=None,
    enabled=None,
):
    """
    Manage the state of upstart managed services.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + reloaded: whether the service should be reloaded
    + command: custom command to pass like: ``/etc/rc.d/<service> <command>``
    + enabled: whether this service should be enabled/disabled on boot

    Enabling/disabling services:
        Upstart jobs define runlevels in their config files - as such there is no way to
        edit/list these without fiddling with the config. So pyinfra simply manages the
        existence of a ``/etc/init/<service>.override`` file, and sets its content to
        "manual" to disable automatic start of services.
    """

    yield from handle_service_control(
        host,
        service,
        host.get_fact(UpstartStatus),
        "initctl {1} {0}",
        running,
        restarted,
        reloaded,
        command,
    )

    # Upstart jobs are setup w/runlevels etc in their config files, so here we just check
    # there's no override file.
    if enabled is True:
        yield from files.file(
            "/etc/init/{0}.override".format(service),
            present=False,
        )

    # Set the override file to "manual" to disable automatic start
    elif enabled is False:
        file = StringIO("manual\n")
        yield from files.put(
            src=file,
            dest="/etc/init/{0}.override".format(service),
        )



import json
import os
from datetime import datetime
from importlib import import_module
from io import IOBase
from os import path
from pathlib import Path
from types import FunctionType, ModuleType
from typing import TYPE_CHECKING

import click
import gevent

from pyinfra import logger, state
from pyinfra.api.command import PyinfraCommand
from pyinfra.api.exceptions import PyinfraError
from pyinfra.api.host import HostData
from pyinfra.api.operation import OperationMeta
from pyinfra.context import ctx_config, ctx_host
from pyinfra.progress import progress_spinner

from .exceptions import CliError, UnexpectedExternalError

if TYPE_CHECKING:
    from pyinfra.api.state import State

# Cache for compiled Python deploy code
PYTHON_CODES = {}


def is_subdir(child, parent):
    child = path.realpath(child)
    parent = path.realpath(parent)
    relative = path.relpath(child, start=parent)
    return not relative.startswith(os.pardir)


def exec_file(filename, return_locals: bool = False, is_deploy_code: bool = False):
    """
    Execute a Python file and optionally return it's attributes as a dict.
    """

    old_current_exec_filename = state.current_exec_filename
    state.current_exec_filename = filename

    if filename not in PYTHON_CODES:
        with open(filename, "r") as f:
            code = f.read()

        code = compile(code, filename, "exec")
        PYTHON_CODES[filename] = code

    # Create some base attributes for our "module"
    data = {
        "__file__": filename,
    }

    # Execute the code with locals/globals going into the dict above
    try:
        exec(PYTHON_CODES[filename], data)
    except Exception as e:
        if isinstance(e, PyinfraError):
            raise
        raise UnexpectedExternalError(e, filename)

    state.current_exec_filename = old_current_exec_filename
    return data


def json_encode(obj):
    # pyinfra types
    if isinstance(obj, HostData):
        return obj.dict()

    if isinstance(obj, PyinfraCommand):
        return repr(obj)

    if isinstance(obj, OperationMeta):
        return repr(obj)

    # Python types
    if isinstance(obj, ModuleType):
        return "Module: {0}".format(obj.__name__)

    if isinstance(obj, FunctionType):
        return "Function: {0}".format(obj.__name__)

    if isinstance(obj, datetime):
        return obj.isoformat()

    if isinstance(obj, IOBase):
        if hasattr(obj, "name"):
            return "File: {0}".format(obj.name)

        if hasattr(obj, "template"):
            return "Template: {0}".format(obj.template)

        obj.seek(0)
        return "In memory file: {0}".format(obj.read())

    if isinstance(obj, Path):
        return str(obj)

    if isinstance(obj, set):
        return sorted(list(obj))

    if isinstance(obj, bytes):
        return obj.decode()

    raise TypeError("Cannot serialize: {0} ({1})".format(type(obj), obj))


def parse_cli_arg(arg):
    if isinstance(arg, list):
        return [parse_cli_arg(a) for a in arg]

    if arg.lower() == "false":
        return False

    if arg.lower() == "true":
        return True

    try:
        return int(arg)
    except (TypeError, ValueError):
        pass

    try:
        return json.loads(arg)
    except ValueError:
        pass

    return arg


def get_operation_and_args(commands):
    operation_name = commands[0]

    # Get the module & operation name
    op_module, op_name = operation_name.rsplit(".", 1)

    # Try to load the requested operation from the main operations package.
    # If that fails, try to load from the user's operations package.
    try:
        op_module = import_module("pyinfra.operations.{0}".format(op_module))
    except ImportError:
        try:
            op_module = import_module(str(op_module))
        except ImportError:
            raise CliError("No such module: {0}".format(op_module))

    op = getattr(op_module, op_name, None)
    if not op:
        raise CliError("No such operation: {0}".format(operation_name))

    # Parse the arguments
    operation_args = commands[1:]

    if len(operation_args) == 1:
        # Check if we're JSON (in which case we expect a list of two items:
        # a list of args and a dict of kwargs).
        try:
            args, kwargs = json.loads(operation_args[0])
            return op, (args or (), kwargs or {})
        except ValueError:
            pass

    args = [parse_cli_arg(arg) for arg in operation_args if "=" not in arg]

    kwargs = {
        key: parse_cli_arg(value)
        for key, value in [arg.split("=", 1) for arg in operation_args if "=" in arg]
    }

    return op, (args, kwargs)


def get_facts_and_args(commands):
    facts = []

    current_fact = None

    for command in commands:
        if "=" in command:
            if not current_fact:
                raise CliError("Invalid fact commands: `{0}`".format(commands))

            key, value = command.split("=", 1)
            current_fact[2][key] = value
            continue

        if current_fact:
            facts.append(current_fact)
            current_fact = None

        if "." not in command:
            raise CliError(f"Invalid fact: `{command}`, should be in the format `module.cls`")

        fact_module, fact_name = command.rsplit(".", 1)
        try:
            fact_module = import_module("pyinfra.facts.{0}".format(fact_module))
        except ImportError:
            try:
                fact_module = import_module(str(fact_module))
            except ImportError:
                raise CliError("No such module: `{0}`".format(fact_module))

        fact_cls = getattr(fact_module, fact_name, None)
        if not fact_cls:
            raise CliError("No such fact: `{0}`".format(command))

        current_fact = (fact_cls, (), {})

    if current_fact:
        facts.append(current_fact)

    return facts


def load_deploy_file(state: "State", filename):
    state.current_deploy_filename = filename

    def load_file(local_host):
        with ctx_config.use(state.config.copy()):
            with ctx_host.use(local_host):
                exec_file(filename)
                logger.info(
                    "{0}{1} {2}".format(
                        local_host.print_prefix,
                        click.style("Ready:", "green"),
                        click.style(filename, bold=True),
                    ),
                )

    greenlet_to_host = {
        state.pool.spawn(load_file, host): host for host in state.inventory.iter_active_hosts()
    }

    with progress_spinner(greenlet_to_host.values()) as progress:
        for greenlet in gevent.iwait(greenlet_to_host.keys()):
            host = greenlet_to_host[greenlet]
            greenlet.get()
            progress(host)



"""
Manage ``choco`` (Chocolatey) packages (https://chocolatey.org).
"""

from pyinfra import host
from pyinfra.api import operation
from pyinfra.facts.choco import ChocoPackages

from .util.packaging import ensure_packages


@operation
def packages(packages=None, present=True, latest=False):
    """
    Add/remove/update ``choco`` packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version

    Versions:
        Package versions can be pinned like gem: ``<pkg>:<version>``.

    **Example:**

    .. code:: python

        # Note: Assumes that 'choco' is installed and
        #       user has Administrator permission.
        choco.packages(
            name="Install Notepad++",
            packages=["notepadplusplus"],
        )
    """

    yield from ensure_packages(
        host,
        packages,
        host.get_fact(ChocoPackages),
        present,
        install_command="choco install -y",
        uninstall_command="choco uninstall -y -x",
        upgrade_command="choco update -y",
        version_join=":",
        latest=latest,
    )


@operation(is_idempotent=False)
def install():
    """
    Install ``choco`` (Chocolatey).
    """

    yield "Set-ExecutionPolicy Bypass -Scope Process -Force ;" "iex ((New-Object System.Net.WebClient).DownloadString" '("https://chocolatey.org/install.ps1"))'  # noqa



"""
The server module takes care of os-level state. Targets POSIX compatibility, tested on
Linux/BSD.
"""

import shlex
from io import StringIO
from itertools import filterfalse, tee
from os import path
from time import sleep

from pyinfra import host, state
from pyinfra.api import FunctionCommand, OperationError, StringCommand, operation
from pyinfra.api.util import try_int
from pyinfra.connectors.util import remove_any_sudo_askpass_file
from pyinfra.facts.files import Directory
from pyinfra.facts.server import (
    Crontab,
    Groups,
    Hostname,
    KernelModules,
    Mounts,
    Os,
    Sysctl,
    Users,
    Which,
)

from . import (
    apk,
    apt,
    brew,
    bsdinit,
    dnf,
    files,
    openrc,
    pacman,
    pkg,
    systemd,
    sysvinit,
    upstart,
    xbps,
    yum,
    zypper,
)
from .util.files import chmod, sed_replace


@operation(is_idempotent=False)
def reboot(delay=10, interval=1, reboot_timeout=300):
    """
    Reboot the server and wait for reconnection.

    + delay: number of seconds to wait before attempting reconnect
    + interval: interval (s) between reconnect attempts
    + reboot_timeout: total time before giving up reconnecting

    **Example:**

    .. code:: python

        server.reboot(
            name="Reboot the server and wait to reconnect",
            delay=60,
            reboot_timeout=600,
        )
    """

    # Remove this now, before we reboot the server - if the reboot fails (expected or
    # not) we'll error if we don't clean this up now. Will simply be re-uploaded if
    # needed later.
    def remove_any_askpass_file(state, host):
        remove_any_sudo_askpass_file(host)

    yield FunctionCommand(remove_any_askpass_file, (), {})

    yield StringCommand("reboot", success_exit_codes=[0, -1])  # -1 being error/disconnected

    def wait_and_reconnect(state, host):  # pragma: no cover
        sleep(delay)
        max_retries = round(reboot_timeout / interval)

        host.connection = None  # remove the connection object
        retries = 0

        while True:
            host.connect(show_errors=False)
            if host.connection:
                break

            if retries > max_retries:
                raise Exception(
                    ("Server did not reboot in time (reboot_timeout={0}s)").format(reboot_timeout),
                )

            sleep(interval)
            retries += 1

    yield FunctionCommand(wait_and_reconnect, (), {})


@operation(is_idempotent=False)
def wait(port: int):
    """
    Waits for a port to come active on the target machine. Requires netstat, checks every
    second.

    + port: port number to wait for

    **Example:**

    .. code:: python

        server.wait(
            name="Wait for webserver to start",
            port=80,
        )
    """

    yield r"""
        while ! (netstat -an | grep LISTEN | grep -e "\.{0}" -e ":{0}"); do
            echo "waiting for port {0}..."
            sleep 1
        done
    """.format(
        port,
    )


@operation(is_idempotent=False)
def shell(commands):
    """
    Run raw shell code on server during a deploy. If the command would
    modify data that would be in a fact, the fact would not be updated
    since facts are only run at the start of a deploy.

    + commands: command or list of commands to execute on the remote server

    **Example:**

    .. code:: python

        server.shell(
            name="Run lxd auto init",
            commands=["lxd init --auto"],
        )
    """

    # Ensure we have a list
    if isinstance(commands, str):
        commands = [commands]

    for command in commands:
        yield command


@operation(is_idempotent=False)
def script(src):
    """
    Upload and execute a local script on the remote host.

    + src: local script filename to upload & execute

    **Example:**

    .. code:: python

        # Note: This assumes there is a file in files/hello.bash locally.
        server.script(
            name="Hello",
            src="files/hello.bash",
        )
    """

    temp_file = state.get_temp_filename(src)
    yield from files.put(src, temp_file)

    yield chmod(temp_file, "+x")
    yield temp_file


@operation(is_idempotent=False)
def script_template(src, **data):
    """
    Generate, upload and execute a local script template on the remote host.

    + src: local script template filename

    **Example:**

    .. code:: python

        # Example showing how to pass python variable to a script template file.
        # The .j2 file can use `{{ some_var }}` to be interpolated.
        # To see output need to run pyinfra with '-v'
        # Note: This assumes there is a file in templates/hello2.bash.j2 locally.
        some_var = 'blah blah blah '
        server.script_template(
            name="Hello from script",
            src="templates/hello2.bash.j2",
            some_var=some_var,
        )
    """

    temp_file = state.get_temp_filename("{0}{1}".format(src, data))
    yield from files.template(src, temp_file, **data)

    yield chmod(temp_file, "+x")
    yield temp_file


@operation
def modprobe(module, present=True, force=False):
    """
    Load/unload kernel modules.

    + module: name of the module to manage
    + present: whether the module should be loaded or not
    + force: whether to force any add/remove modules

    **Example:**

    .. code:: python

        server.modprobe(
            name="Silly example for modprobe",
            module="floppy",
        )
    """
    list_value = [module] if isinstance(module, str) else module

    # NOTE: https://docs.python.org/3/library/itertools.html#itertools-recipes
    def partition(predicate, iterable):
        t1, t2 = tee(iterable)
        return list(filter(predicate, t2)), list(filterfalse(predicate, t1))

    modules = host.get_fact(KernelModules)
    present_mods, missing_mods = partition(lambda mod: mod in modules, list_value)

    args = ""
    if force:
        args = " -f"

    # Module is loaded and we don't want it?
    if not present and present_mods:
        yield "modprobe{0} -r -a {1}".format(args, " ".join(present_mods))
        for mod in present_mods:
            modules.pop(mod)

    # Module isn't loaded and we want it?
    elif present and missing_mods:
        yield "modprobe{0} -a {1}".format(args, " ".join(missing_mods))
        for mod in missing_mods:
            modules[mod] = {}

    else:
        host.noop(
            "{0} {1} {2} {3}".format(
                "modules" if len(list_value) > 1 else "module",
                "/".join(list_value),
                "are" if len(list_value) > 1 else "is",
                "loaded" if present else "not loaded",
            ),
        )


@operation
def mount(
    path,
    mounted=True,
    options=None,
    # TODO: do we want to manage fstab here?
    # update_fstab=False, device=None, fs_type=None,
):
    """
    Manage mounted filesystems.

    + path: the path of the mounted filesystem
    + mounted: whether the filesystem should be mounted
    + options: the mount options

    Options:
        If the currently mounted filesystem does not have all of the provided
        options it will be remounted with the options provided.

    ``/etc/fstab``:
        This operation does not attempt to modify the on disk fstab file - for
        that you should use the `files.line operation <./files.html#files-line>`_.
    """

    options = options or []
    options_string = ",".join(options)

    mounts = host.get_fact(Mounts)
    is_mounted = path in mounts

    # Want mount but don't have?
    if mounted and not is_mounted:
        yield "mount{0} {1}".format(
            " -o {0}".format(options_string) if options_string else "",
            path,
        )
        mounts[path] = {"options": options}

    # Want no mount but mounted?
    elif mounted is False and is_mounted:
        yield "umount {0}".format(path)
        mounts.pop(path)

    # Want mount and is mounted! Check the options
    elif is_mounted and mounted and options:
        mounted_options = mounts[path]["options"]
        needed_options = set(options) - set(mounted_options)
        if needed_options:
            yield "mount -o remount,{0} {1}".format(options_string, path)
            mounts[path]["options"] = options

    else:
        host.noop(
            "filesystem {0} is {1}".format(
                path,
                "mounted" if mounted else "not mounted",
            ),
        )


@operation
def hostname(hostname, hostname_file=None):
    """
    Set the system hostname using ``hostnamectl`` or ``hostname`` on older systems.

    + hostname: the hostname that should be set
    + hostname_file: the file that permanently sets the hostname

    Hostname file:
        The hostname file only matters no systems that do not have ``hostnamectl``,
        which is part of ``systemd``.

        By default pyinfra will auto detect this by targeting ``/etc/hostname``
        on Linux and ``/etc/myname`` on OpenBSD.

        To completely disable writing the hostname file, set ``hostname_file=False``.

    **Example:**

    .. code:: python

        server.hostname(
            name="Set the hostname",
            hostname="server1.example.com",
        )
    """

    current_hostname = host.get_fact(Hostname)

    if host.get_fact(Which, command="hostnamectl"):
        if current_hostname != hostname:
            yield "hostnamectl set-hostname {0}".format(hostname)
            host.create_fact(Hostname, data=hostname)
        else:
            host.noop("hostname is set")
        return

    if hostname_file is None:
        os = host.get_fact(Os)

        if os == "Linux":
            hostname_file = "/etc/hostname"
        elif os == "OpenBSD":
            hostname_file = "/etc/myname"

    if current_hostname != hostname:
        yield "hostname {0}".format(hostname)
        host.create_fact(Hostname, data=hostname)
    else:
        host.noop("hostname is set")

    if hostname_file:
        # Create a whole new hostname file
        file = StringIO("{0}\n".format(hostname))

        # And ensure it exists
        yield from files.put(file, hostname_file)


@operation
def sysctl(
    key,
    value,
    persist=False,
    persist_file="/etc/sysctl.conf",
):
    """
    Edit sysctl configuration.

    + key: name of the sysctl setting to ensure
    + value: the value or list of values the sysctl should be
    + persist: whether to write this sysctl to the config
    + persist_file: file to write the sysctl to persist on reboot

    **Example:**

    .. code:: python

        server.sysctl(
            name="Change the fs.file-max value",
            key="fs.file-max",
            value=100000,
            persist=True,
        )
    """

    string_value = " ".join(["{0}".format(v) for v in value]) if isinstance(value, list) else value

    value = [try_int(v) for v in value] if isinstance(value, list) else try_int(value)

    existing_sysctls = host.get_fact(Sysctl)

    existing_value = existing_sysctls.get(key)
    if not existing_value or existing_value != value:
        yield "sysctl {0}='{1}'".format(key, string_value)
        existing_sysctls[key] = value
    else:
        host.noop("sysctl {0} is set to {1}".format(key, string_value))

    if persist:
        yield from files.line(
            path=persist_file,
            line="{0}[[:space:]]*=[[:space:]]*{1}".format(key, string_value),
            replace="{0} = {1}".format(key, string_value),
        )


@operation
def service(
    service,
    running=True,
    restarted=False,
    reloaded=False,
    command=None,
    enabled=None,
):
    """
    Manage the state of services. This command checks for the presence of all the
    Linux init systems ``pyinfra`` can handle and executes the relevant operation.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + reloaded: whether the service should be reloaded
    + command: custom command execute
    + enabled: whether this service should be enabled/disabled on boot

    **Example:**

    .. code:: python

        server.service(
            name="Enable open-vm-tools service",
            service="open-vm-tools",
            enabled=True,
        )
    """

    if host.get_fact(Which, command="systemctl"):
        service_operation = systemd.service

    elif host.get_fact(Which, command="rc-service"):
        service_operation = openrc.service

    elif host.get_fact(Which, command="initctl"):
        service_operation = upstart.service

    elif host.get_fact(Directory, path="/etc/init.d"):
        service_operation = sysvinit.service

    elif host.get_fact(Directory, path="/etc/rc.d"):
        service_operation = bsdinit.service

    else:
        raise OperationError(
            ("No init system found " "(no systemctl, initctl, /etc/init.d or /etc/rc.d found)"),
        )

    yield from service_operation(
        service,
        running=running,
        restarted=restarted,
        reloaded=reloaded,
        command=command,
        enabled=enabled,
    )


@operation
def packages(
    packages,
    present=True,
):
    """
    Add or remove system packages. This command checks for the presence of all the
    system package managers ``pyinfra`` can handle and executes the relevant operation.

    + packages: list of packages to ensure
    + present: whether the packages should be installed

    **Example:**

    .. code:: python

        server.packages(
            name="Install Vim and vimpager",
            packages=["vimpager", "vim"],
        )
    """

    # TODO: improve this - use LinuxDistribution fact + mapping with fallback below?
    # Here to be preferred on openSUSE which also provides aptitude
    # See: https://github.com/Fizzadar/pyinfra/issues/799
    if host.get_fact(Which, command="zypper"):
        package_operation = zypper.packages

    elif host.get_fact(Which, command="apk"):
        package_operation = apk.packages

    elif host.get_fact(Which, command="apt"):
        package_operation = apt.packages

    elif host.get_fact(Which, command="brew"):
        package_operation = brew.packages

    elif host.get_fact(Which, command="dnf"):
        package_operation = dnf.packages

    elif host.get_fact(Which, command="pacman"):
        package_operation = pacman.packages

    elif host.get_fact(Which, command="xbps"):
        package_operation = xbps.packages

    elif host.get_fact(Which, command="yum"):
        package_operation = yum.packages

    elif host.get_fact(Which, command="pkg") or host.get_fact(Which, command="pkg_add"):
        package_operation = pkg.packages

    else:
        raise OperationError(
            (
                "No system package manager found "
                "(no apk, apt, brew, dnf, pacman, pkg, xbps, yum or zypper found)"
            ),
        )

    yield from package_operation(packages=packages, present=present)


@operation
def crontab(
    command,
    present=True,
    user=None,
    cron_name=None,
    minute="*",
    hour="*",
    month="*",
    day_of_week="*",
    day_of_month="*",
    special_time=None,
    interpolate_variables=False,
):
    """
    Add/remove/update crontab entries.

    + command: the command for the cron
    + present: whether this cron command should exist
    + user: the user whose crontab to manage
    + cron_name: name the cronjob so future changes to the command will overwrite
    + minute: which minutes to execute the cron
    + hour: which hours to execute the cron
    + month: which months to execute the cron
    + day_of_week: which day of the week to execute the cron
    + day_of_month: which day of the month to execute the cron
    + special_time: cron "nickname" time (@reboot, @daily, etc), overrides others
    + interpolate_variables: whether to interpolate variables in ``command``

    Cron commands:
        Unless ``name`` is specified the command is used to identify crontab entries.
        This means commands must be unique within a given users crontab. If you require
        multiple identical commands, provide a different name argument for each.

    Special times:
        When provided, ``special_time`` will be used instead of any values passed in
        for ``minute``/``hour``/``month``/``day_of_week``/``day_of_month``.

    **Example:**

    .. code:: python

        # simple example for a crontab
        server.crontab(
            name="Backup /etc weekly",
            command="/bin/tar cf /tmp/etc_bup.tar /etc",
            name="backup_etc",
            day_of_week=0,
            hour=1,
            minute=0,
        )
    """

    def comma_sep(value):
        if isinstance(value, (list, tuple)):
            return ",".join("{0}".format(v) for v in value)
        return value

    minute = comma_sep(minute)
    hour = comma_sep(hour)
    month = comma_sep(month)
    day_of_week = comma_sep(day_of_week)
    day_of_month = comma_sep(day_of_month)

    crontab = host.get_fact(Crontab, user=user)
    name_comment = "# pyinfra-name={0}".format(cron_name)

    existing_crontab = crontab.get(command)
    existing_crontab_command = command
    existing_crontab_match = command

    if not existing_crontab and cron_name:  # find the crontab by name if provided
        for cmd, details in crontab.items():
            if name_comment in details["comments"]:
                existing_crontab = details
                existing_crontab_match = cmd
                existing_crontab_command = cmd

    exists = existing_crontab is not None

    edit_commands = []
    temp_filename = state.get_temp_filename()

    if special_time:
        new_crontab_line = "{0} {1}".format(special_time, command)
    else:
        new_crontab_line = "{minute} {hour} {day_of_month} {month} {day_of_week} {command}".format(
            minute=minute,
            hour=hour,
            day_of_month=day_of_month,
            month=month,
            day_of_week=day_of_week,
            command=command,
        )

    existing_crontab_match = ".*{0}.*".format(existing_crontab_match)

    # Don't want the cron and it does exist? Remove the line
    if not present and exists:
        edit_commands.append(
            sed_replace(
                temp_filename,
                existing_crontab_match,
                "",
                interpolate_variables=interpolate_variables,
            ),
        )

    # Want the cron but it doesn't exist? Append the line
    elif present and not exists:
        if cron_name:
            if crontab:  # append a blank line if cron entries already exist
                edit_commands.append("echo '' >> {0}".format(temp_filename))
            edit_commands.append(
                "echo {0} >> {1}".format(
                    shlex.quote(name_comment),
                    temp_filename,
                ),
            )

        edit_commands.append(
            "echo {0} >> {1}".format(
                shlex.quote(new_crontab_line),
                temp_filename,
            ),
        )

    # We have the cron and it exists, do it's details? If not, replace the line
    elif present and exists:
        if any(
            (
                special_time != existing_crontab.get("special_time"),
                minute != existing_crontab.get("minute"),
                hour != existing_crontab.get("hour"),
                month != existing_crontab.get("month"),
                day_of_week != existing_crontab.get("day_of_week"),
                day_of_month != existing_crontab.get("day_of_month"),
                existing_crontab_command != command,
            ),
        ):
            edit_commands.append(
                sed_replace(
                    temp_filename,
                    existing_crontab_match,
                    new_crontab_line,
                    interpolate_variables=interpolate_variables,
                ),
            )

    if edit_commands:
        crontab_args = []
        if user:
            crontab_args.append("-u {0}".format(user))

        # List the crontab into a temporary file if it exists
        if crontab:
            yield "crontab -l {0} > {1}".format(" ".join(crontab_args), temp_filename)

        # Now yield any edits
        for edit_command in edit_commands:
            yield edit_command

        # Finally, use the tempfile to write a new crontab
        yield "crontab {0} {1}".format(" ".join(crontab_args), temp_filename)

        # Update the crontab fact
        if present:
            crontab[command] = {
                "special_time": special_time,
                "minute": minute,
                "hour": hour,
                "month": month,
                "day_of_week": day_of_week,
                "day_of_month": day_of_month,
                "comments": [cron_name] if cron_name else [],
            }
        else:
            crontab.pop(command)
    else:
        host.noop(
            "crontab {0} {1}".format(
                command,
                "exists" if present else "does not exist",
            ),
        )


@operation
def group(group, present=True, system=False, gid=None):
    """
    Add/remove system groups.

    + group: name of the group to ensure
    + present: whether the group should be present or not
    + system: whether to create a system group

    System users:
        System users don't exist on BSD, so the argument is ignored for BSD targets.

    **Examples:**

    .. code:: python

        server.group(
            name="Create docker group",
            group="docker",
        )

        # multiple groups
        for group in ["wheel", "lusers"]:
            server.group(
                name=f"Create the group {group}",
                group=group,
            )
    """

    groups = host.get_fact(Groups)
    is_present = group in groups

    # Group exists but we don't want them?
    if not present and is_present:
        yield "groupdel {0}".format(group)
        groups.remove(group)

    # Group doesn't exist and we want it?
    elif present and not is_present:
        args = []

        # BSD doesn't do system users
        if system and "BSD" not in host.get_fact(Os):
            args.append("-r")

        args.append(group)

        if gid:
            args.append("--gid {0}".format(gid))

        # Groups are often added by other operations (package installs), so check
        # for the group at runtime before adding.
        yield "grep '^{0}:' /etc/group || groupadd {1}".format(
            group,
            " ".join(args),
        )
        groups.append(group)


@operation
def user_authorized_keys(
    user,
    public_keys,
    group=None,
    delete_keys=False,
    authorized_key_directory=None,
    authorized_key_filename=None,
):
    """
    Manage `authorized_keys` of system users.

    + user: name of the user to ensure
    + public_keys: list of public keys to attach to this user, ``home`` must be specified
    + group: the users primary group
    + delete_keys: whether to remove any keys not specified in ``public_keys``

    Public keys:
        These can be provided as strings containing the public key or as a path to
        a public key file which ``pyinfra`` will read.

    **Examples:**

    .. code:: python

        server.user_authorized_keys(
            name="Ensure user has a public key",
            user="kevin",
            public_keys=["ed25519..."],
        )
    """
    if not authorized_key_directory:
        authorized_key_directory = f"/home/{user}/.ssh/"
    if not authorized_key_filename:
        authorized_key_filename = "authorized_keys"

    if isinstance(public_keys, str):
        public_keys = [public_keys]

    def read_any_pub_key_file(key):
        try_path = key
        if state.cwd:
            try_path = path.join(state.cwd, key)

        if path.exists(try_path):
            with open(try_path, "r") as f:
                return f.read()

        return key

    public_keys = list(map(read_any_pub_key_file, public_keys))

    # Ensure .ssh directory
    # note that this always outputs commands unless the SSH user has access to the
    # authorized_keys file, ie the SSH user is the user defined in this function
    yield from files.directory(
        authorized_key_directory,
        user=user,
        group=group or user,
        mode=700,
    )

    authorized_key_file = f"{authorized_key_directory}/{authorized_key_filename}"

    if delete_keys:
        # Create a whole new authorized_keys file
        keys_file = StringIO(
            "{0}\n".format(
                "\n".join(public_keys),
            ),
        )

        # And ensure it exists
        yield from files.put(
            src=keys_file,
            dest=authorized_key_file,
            user=user,
            group=group or user,
            mode=600,
        )

    else:
        # Ensure authorized_keys exists
        yield from files.file(
            path=authorized_key_file,
            user=user,
            group=group or user,
            mode=600,
        )

        # And every public key is present
        for key in public_keys:
            yield from files.line(path=authorized_key_file, line=key)


@operation
def user(
    user,
    present=True,
    home=None,
    shell=None,
    group=None,
    groups=None,
    public_keys=None,
    delete_keys=False,
    ensure_home=True,
    create_home=False,
    system=False,
    uid=None,
    comment=None,
    add_deploy_dir=True,
    unique=True,
):
    """
    Add/remove/update system users & their ssh `authorized_keys`.

    + user: name of the user to ensure
    + present: whether this user should exist
    + home: the users home directory
    + shell: the users shell
    + group: the users primary group
    + groups: the users secondary groups
    + public_keys: list of public keys to attach to this user, ``home`` must be specified
    + delete_keys: whether to remove any keys not specified in ``public_keys``
    + ensure_home: whether to ensure the ``home`` directory exists
    + create_home: whether to new user create home directories from the system skeleton
    + system: whether to create a system account
    + comment: the user GECOS comment
    + add_deploy_dir: any public_key filenames are relative to the deploy directory
    + unique: prevent creating users with duplicate UID

    Home directory:
        When ``ensure_home`` or ``public_keys`` are provided, ``home`` defaults to
        ``/home/{name}``. When ``create_home`` is ``True`` any newly created users
        will be created with the ``-m`` flag to build a new home directory from the
        systems skeleton directory.

    Public keys:
        These can be provided as strings containing the public key or as a path to
        a public key file which ``pyinfra`` will read.

    **Examples:**

    .. code:: python

        server.user(
            name="Ensure user is removed",
            user="kevin",
            present=False,
        )

        server.user(
            name="Ensure myweb user exists",
            user="myweb",
            shell="/bin/bash",
        )

        # multiple users
        for user in ["kevin", "bob"]:
            server.user(
                name=f"Ensure user {user} is removed",
                user=user,
                present=False,
            )
    """

    users = host.get_fact(Users)
    existing_groups = host.get_fact(Groups)
    existing_user = users.get(user)

    if groups is None:
        groups = []

    if home is None:
        home = "/home/{0}".format(user)
        if existing_user:
            home = existing_user.get("home", home)

    # User not wanted?
    if not present:
        if existing_user:
            yield "userdel {0}".format(user)
            users.pop(user)
        return

    # User doesn't exist but we want them?
    if present and existing_user is None:
        # Fix the case where a group of the same name already exists, tell useradd to use this
        # group rather than failing trying to create it.
        if not group and user in existing_groups:
            group = user

        # Create the user w/home/shell
        args = []

        if home:
            args.append("-d {0}".format(home))

        if shell:
            args.append("-s {0}".format(shell))

        if group:
            args.append("-g {0}".format(group))

        if groups:
            args.append("-G {0}".format(",".join(groups)))

        if system and "BSD" not in host.get_fact(Os):
            args.append("-r")

        if uid:
            args.append("--uid {0}".format(uid))

        if comment:
            args.append("-c '{0}'".format(comment))

        if not unique:
            args.append("-o")

        if create_home:
            args.append("-m")

        # Users are often added by other operations (package installs), so check
        # for the user at runtime before adding.
        yield "grep '^{1}:' /etc/passwd || useradd {0} {1}".format(
            " ".join(args),
            user,
        )
        users[user] = {
            "comment": comment,
            "home": home,
            "shell": shell,
            "group": group,
            "groups": groups,
        }

    # User exists and we want them, check home/shell/keys
    else:
        args = []

        # Check homedir
        if home and existing_user["home"] != home:
            args.append("-d {0}".format(home))

        # Check shell
        if shell and existing_user["shell"] != shell:
            args.append("-s {0}".format(shell))

        # Check primary group
        if group and existing_user["group"] != group:
            args.append("-g {0}".format(group))

        # Check secondary groups, if defined
        if groups and set(existing_user["groups"]) != set(groups):
            args.append("-G {0}".format(",".join(groups)))

        if comment and existing_user["comment"] != comment:
            args.append("-c '{0}'".format(comment))

        # Need to mod the user?
        if args:
            yield "usermod {0} {1}".format(" ".join(args), user)
            if comment:
                existing_user["comment"] = comment
            if home:
                existing_user["home"] = home
            if shell:
                existing_user["shell"] = shell
            if group:
                existing_user["group"] = group
            if groups:
                existing_user["groups"] = groups

    # Ensure home directory ownership
    if ensure_home:
        yield from files.directory(
            home,
            user=user,
            group=group or user,
            # Don't fail if the home directory exists as a link
            _no_fail_on_link=True,
        )

    # Add SSH keys
    if public_keys is not None:
        yield from user_authorized_keys(
            user,
            public_keys,
            group=group,
            delete_keys=delete_keys,
            authorized_key_directory="{0}/.ssh".format(home),
            authorized_key_filename=None,
        )



import shlex
from subprocess import PIPE
from unittest import TestCase
from unittest.mock import MagicMock, mock_open, patch

from pyinfra.api import Config, State
from pyinfra.api.connect import connect_all
from pyinfra.api.exceptions import InventoryError, PyinfraError
from pyinfra.connectors.util import make_unix_command

from ..util import make_inventory


def fake_docker_shell(command, splitlines=None):
    if command == "docker run -d not-an-image tail -f /dev/null":
        return ["containerid"]

    elif command == "docker commit containerid":
        return ["sha256:blahsomerandomstringdata"]

    elif command == "docker rm -f containerid":
        return []

    raise PyinfraError("Invalid command: {0}".format(command))


@patch("pyinfra.connectors.docker.local.shell", fake_docker_shell)
@patch("pyinfra.connectors.docker.mkstemp", lambda: (None, "__tempfile__"))
@patch("pyinfra.connectors.docker.os.remove", lambda f: None)
@patch("pyinfra.connectors.docker.os.close", lambda f: None)
@patch("pyinfra.connectors.docker.open", mock_open(read_data="test!"), create=True)
@patch("pyinfra.api.util.open", mock_open(read_data="test!"), create=True)
class TestDockerConnector(TestCase):
    def setUp(self):
        self.fake_popen_patch = patch("pyinfra.connectors.util.Popen")
        self.fake_popen_mock = self.fake_popen_patch.start()

    def tearDown(self):
        self.fake_popen_patch.stop()

    def test_missing_image(self):
        with self.assertRaises(InventoryError):
            make_inventory(hosts=("@docker",))

    def test_user_provided_container_id(self):
        inventory = make_inventory(
            hosts=(("@docker/not-an-image", {"docker_container_id": "abc"}),),
        )
        State(inventory, Config())
        host = inventory.get_host("@docker/not-an-image")
        host.connect()
        assert host.data.docker_container_id == "abc"

    def test_connect_all(self):
        inventory = make_inventory(hosts=("@docker/not-an-image",))
        state = State(inventory, Config())
        connect_all(state)
        assert len(state.active_hosts) == 1

    def test_connect_all_error(self):
        inventory = make_inventory(hosts=("@docker/a-broken-image",))
        state = State(inventory, Config())

        with self.assertRaises(PyinfraError):
            connect_all(state)

    def test_connect_disconnect_host(self):
        inventory = make_inventory(hosts=("@docker/not-an-image",))
        state = State(inventory, Config())
        host = inventory.get_host("@docker/not-an-image")
        host.connect(reason=True)
        assert len(state.active_hosts) == 0
        host.disconnect()

    def test_run_shell_command(self):
        inventory = make_inventory(hosts=("@docker/not-an-image",))
        State(inventory, Config())

        command = "echo hi"
        self.fake_popen_mock().returncode = 0

        host = inventory.get_host("@docker/not-an-image")
        host.connect()
        out = host.run_shell_command(
            command,
            stdin="hello",
            get_pty=True,
            print_output=True,
        )
        assert len(out) == 3
        assert out[0] is True

        command = make_unix_command(command).get_raw_value()
        command = shlex.quote(command)
        docker_command = "docker exec -it containerid sh -c {0}".format(command)
        shell_command = make_unix_command(docker_command).get_raw_value()

        self.fake_popen_mock.assert_called_with(
            shell_command,
            shell=True,
            stdout=PIPE,
            stderr=PIPE,
            stdin=PIPE,
        )

    def test_run_shell_command_success_exit_codes(self):
        inventory = make_inventory(hosts=("@docker/not-an-image",))
        State(inventory, Config())

        command = "echo hi"
        self.fake_popen_mock().returncode = 1

        host = inventory.get_host("@docker/not-an-image")
        host.connect()
        out = host.run_shell_command(command, success_exit_codes=[1])
        assert out[0] is True

    def test_run_shell_command_error(self):
        inventory = make_inventory(hosts=("@docker/not-an-image",))
        state = State(inventory, Config())

        command = "echo hi"
        self.fake_popen_mock().returncode = 1

        host = inventory.get_host("@docker/not-an-image")
        host.connect(state)
        out = host.run_shell_command(command)
        assert out[0] is False

    def test_put_file(self):
        inventory = make_inventory(hosts=("@docker/not-an-image",))
        State(inventory, Config())

        host = inventory.get_host("@docker/not-an-image")
        host.connect()

        fake_process = MagicMock(returncode=0)
        self.fake_popen_mock.return_value = fake_process

        host.put_file("not-a-file", "not-another-file", print_output=True)

        self.fake_popen_mock.assert_called_with(
            "sh -c 'docker cp __tempfile__ containerid:not-another-file'",
            shell=True,
            stdout=PIPE,
            stderr=PIPE,
            stdin=PIPE,
        )

    def test_put_file_error(self):
        inventory = make_inventory(hosts=("@docker/not-an-image",))
        State(inventory, Config())

        host = inventory.get_host("@docker/not-an-image")
        host.connect()

        fake_process = MagicMock(returncode=1)
        self.fake_popen_mock.return_value = fake_process

        with self.assertRaises(IOError):
            host.put_file("not-a-file", "not-another-file", print_output=True)

    def test_get_file(self):
        inventory = make_inventory(hosts=("@docker/not-an-image",))
        State(inventory, Config())

        host = inventory.get_host("@docker/not-an-image")
        host.connect()

        fake_process = MagicMock(returncode=0)
        self.fake_popen_mock.return_value = fake_process

        host.get_file("not-a-file", "not-another-file", print_output=True)

        self.fake_popen_mock.assert_called_with(
            "sh -c 'docker cp containerid:not-a-file __tempfile__'",
            shell=True,
            stdout=PIPE,
            stderr=PIPE,
            stdin=PIPE,
        )

    def test_get_file_error(self):
        inventory = make_inventory(hosts=("@docker/not-an-image",))
        State(inventory, Config())

        host = inventory.get_host("@docker/not-an-image")
        host.connect()

        fake_process = MagicMock(returncode=1)
        self.fake_popen_mock.return_value = fake_process

        with self.assertRaises(IOError):
            host.get_file("not-a-file", "not-another-file", print_output=True)



from os import path

from pkg_resources import Requirement, ResolutionError, parse_version, require

from pyinfra import __version__, state

from .exceptions import PyinfraError


class ConfigDefaults:
    # % of hosts which have to fail for all operations to stop
    FAIL_PERCENT = None
    # Seconds to timeout SSH connections
    CONNECT_TIMEOUT = 10
    # Temporary directory (on the remote side) to use for caching any files/downloads
    TEMP_DIR = "/tmp"
    # Gevent pool size (defaults to #of target hosts)
    PARALLEL = 0
    # Specify the required pyinfra version (using PEP 440 setuptools specifier)
    REQUIRE_PYINFRA_VERSION = None
    # Specify any required packages (either using PEP 440 or a requirements file)
    # Note: this can also include pyinfra potentially replacing REQUIRE_PYINFRA_VERSION
    REQUIRE_PACKAGES = None
    # All these can be overridden inside individual operation calls:
    # Switch to this user (from ssh_user) using su before executing operations
    SU_USER = None
    USE_SU_LOGIN = False
    SU_SHELL = None
    PRESERVE_SU_ENV = False
    # Use sudo and optional user
    SUDO = False
    SUDO_USER = None
    PRESERVE_SUDO_ENV = False
    USE_SUDO_LOGIN = False
    USE_SUDO_PASSWORD = False
    # Use doas and optional user
    DOAS = False
    DOAS_USER = None
    # Use doas and optional user
    DOAS = False
    DOAS_USER = None
    # Only show errors but don't count as failure
    IGNORE_ERRORS = False
    # Shell to use to execute commands
    SHELL = "sh"


config_defaults = {key: value for key, value in ConfigDefaults.__dict__.items() if key.isupper()}


def check_pyinfra_version(version: str):
    if not version:
        return

    running_version = parse_version(__version__)
    required_versions = Requirement.parse(
        "pyinfra{0}".format(version),
    )

    if running_version not in required_versions:
        raise PyinfraError(
            ("pyinfra version requirement not met " "(requires {0}, running {1})").format(
                version,
                __version__,
            ),
        )


def check_require_packages(requirements_config):
    if not requirements_config:
        return

    if isinstance(requirements_config, (list, tuple)):
        requirements = requirements_config
    else:
        with open(path.join(state.cwd, requirements_config)) as f:
            requirements = [line.split("#egg=")[-1] for line in f.read().splitlines()]

    try:
        require(requirements)
    except ResolutionError as e:
        raise PyinfraError(
            "Deploy requirements ({0}) not met: {1}".format(
                requirements_config,
                e,
            ),
        )


config_checkers = {
    "REQUIRE_PYINFRA_VERSION": check_pyinfra_version,
    "REQUIRE_PACKAGES": check_require_packages,
}


class Config(ConfigDefaults):
    """
    The default/base configuration options for a pyinfra deploy.
    """

    def __init__(self, **kwargs):
        # Always apply some env
        env = kwargs.pop("ENV", {})
        self.ENV = env

        config = config_defaults.copy()
        config.update(kwargs)

        for key, value in config.items():
            setattr(self, key, value)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)

        checker = config_checkers.get(key)
        if checker:
            checker(value)

    def get_current_state(self):
        return [(key, getattr(self, key)) for key in config_defaults.keys()]

    def set_current_state(self, config_state):
        for key, value in config_state:
            setattr(self, key, value)

    def lock_current_state(self):
        self._locked_config = self.get_current_state()

    def reset_locked_state(self):
        self.set_current_state(self._locked_config)

    def copy(self) -> "Config":
        return Config(**dict(self.get_current_state()))
